/*
 * File: agent.h
 * Author: Suman Banerjee <suman@cs.umd.edu>
 * Date: 15th February, 2002
 * Terms: GPL
 *
 * NICE Application Layer Multicast
 */

#ifndef _AGENT_H_
#define _AGENT_H_

#include <map>
#include <string.h>
#include "app-packet.h"
#include <sys/time.h>
#include "timer.h"

// change this back
#define LOG_CLUSTER_INFO_DETAILED

#define UNDEFINED_AGENT_ID 100000
//class Node;

typedef void (*SOCKETHANDLER)(char*, int);
typedef map<int, SOCKETHANDLER> SocketHandlerMap;

class Agent;

enum SourceDistributionType {
  DIST_CONSTANT,
  DIST_UNIFORM,
  DIST_EXPONENTIAL
};

enum AgentType {

  AGENT_NONE,
  AGENT_ROUTING,
  AGENT_APPLICATION_RP,
  AGENT_APPLICATION_MC,
  AGENT_APPLICATION_BSE,
  AGENT_APPLICATION_COOP
};

enum AgentEventType {
  AGENT_START,
  AGENT_STOP,
  AGENT_SOURCE
};

struct StructPktInfo {

  int gr_size;
  int curr_size;
  int* members;
  int lid;
  int leader;
  int my_id;

  StructPktInfo() {
    members = NULL;
    reset();
  }
  ~StructPktInfo() {
    delete [] members;
  }
  void reset() {
    delete [] members;
    members = NULL;
    gr_size = curr_size = my_id = 0;
  }
  void set_size(int size) {
    gr_size = size;
    members = new int[size];
  }
  void add_member(int id) {
    members[curr_size++] = id;
  }
};

struct AgentInfo {

  int agent_id;
  //  int node_id;
  struct sockaddr_in agent_addr; 
  AgentInfo () {}
  AgentInfo (int aid, struct sockaddr_in sa) { 
    agent_id = aid;
    memcpy(& agent_addr, &sa, sizeof(struct sockaddr_in));
  }
};

struct AgentEvent {
  AgentEventType t;
  SourceDistributionType sdt;
  int source_burst;
  double source_gap;

  AgentEvent () { ; }
  AgentEvent (AgentEventType type) { t = type; }
};

class DataTimer : public Timer {

  public :
    Agent *a;
  DataTimer (void) : Timer() { a = NULL;};
  DataTimer (Agent *A) : Timer() { a = A;};
  void EventHandler (bool);
};



class Agent {

protected:
  /* This is some stuff used for sending packets to an
     agent to visualize the structure of the multicast 
     group. --Ryan
  */
  int drawer_sock;
  sockaddr_in drawer_addr;
  struct StructPktInfo pkt_info;

public :

  AgentType t;
  int id;
  int index;
  //  Node *n;
  //  struct sockaddr_in udp_send_agent_addr;
  struct sockaddr_in udp_recv_agent_addr;
  int udp_sock; // for sending/receiving 
  fd_set socket_set; // all listening sockets
  int num_sockets;
  int max_socket_fd;
  bool started;
  struct AgentEvent data_traffic_params;
  DataTimer dt;
  SocketHandlerMap socket_map;

  Agent (void) {};
  Agent (int Id, int Index);
  virtual ~Agent (void) {};

  virtual void init (int Id, int Index, char* drawer = NULL, 
		     unsigned short dport = 0);
  virtual void start (void);
  virtual void stop (void);
  // Some functions to set up the packets we will send containing cluster
  // information. -- Ryan
  virtual int struct_pkt_reset();
  virtual int struct_pkt_set_size(int size);
  virtual int struct_pkt_add_member(int id);
  virtual int struct_pkt_set_lid(int lid);
  virtual int struct_pkt_set_ldr(int ldr);
  virtual int send_struct_pkt();

  void add_listening_fd(int fd, SOCKETHANDLER callback);
  void remove_listening_fd(int fd);

  //  void send_pkt (Packet *p, int dst_ag, struct sockaddr_in dst_ag_addr);
  void send_pkt (char * payload, int len, int dst_ag, struct sockaddr_in dst_ag_addr);
  void init_data_traffic_params(SourceDistributionType sdt, int burst_size, double burst_gap, long ticks_from_now);
  int generic_rx_pkt_handler (Packet *p);
  virtual int specific_rx_pkt_handler (Packet *p); 
  /* This is called by generic */

  //  void generic_send_data_pkt (SourceDistributionType sdt, int burst_size, double burst_gap) { return; };
  void generic_send_data_pkt (char* payload, int data_len);
  virtual void specific_send_data_pkt (char* payload, int data_len);

  void EventHandler (AgentEvent *ae);
  virtual int flatten_apppacket(char * payload, AppPacket * ap, int dst_ag) { return 0; };
  virtual AppPacket * reclaim_apppacket(char * recv_buffer, int len) { return NULL; };

protected:
  void put_in_pkt(char* pkt, int value, int index);
};

void init_random_for_pkt_sources (unsigned int seed);
long int get_pkt_source_random (void);
double my_clock();

#endif 
