/*
 * File: app-packet.h
 * Author: Suman Banerjee <suman@cs.umd.edu>
 * Date: 15th February, 2002
 * Terms: GPL
 *
 * NICE Application Layer Multicast
 */

#ifndef _APP_PACKET_H_
#define _APP_PACKET_H_

#include "packet.h"

#define LOWER_K 2
#define UPPER_2K (2*LOWER_K)
#define UPPER_3K (3*LOWER_K)
#define MAX_RET_COUNT 2049

//PRM
#define		MAX_BITMAP		1024
//PRM

enum AppPacketSubType {

  JOIN_QUERY,
  JOIN_FORWARD,
  JOIN_RESPONSE,
  CLUSTER_REFRESH,
  CLUSTER_MERGE,
  PING_QUERY,
  PING_RESPONSE,
  PACKET_DATA,
  PACKET_DATA_ACK,
  PACKET_UNDEFINED,

  //PRM
	QUERY_RANDOM,
	RESPONSE_RANDOM,
  	RETRANSMIT_REQUEST
  //REQUEST_ADOPT 
  //PRM
};


struct PacketAgentInfo {

  int agent_id;
  //int node_id;
  struct sockaddr_in agent_addr;
};

struct const_join_query_pkt {
 int q_lid; /* The queried layer id. -1 for query to BSE */
 int attach; /* If this an attach request */
  //long send_time; /* Time when the packet was sent */
 
  double send_time;
  struct sockaddr_in reply_addr; 
};

struct join_query_pkt {

 int q_lid; /* The queried layer id. -1 for query to BSE */
 bool attach; /* If this an attach request */
 double send_time; /* Time when the packet was sent */
 struct sockaddr_in reply_addr; 
};

struct const_join_forward_pkt {

 PacketAgentInfo src_ag;
 int q_lid; /* The queried layer id. -1 for query to BSE */
 int attach;
 PacketAgentInfo original_dst; /* This was the agent to which the join request
		      * was actually sent to */
  //long send_time; /* Time when the packet was sent */
  double send_time;
};

struct join_forward_pkt {

 PacketAgentInfo src_ag;
 int q_lid; /* The queried layer id. -1 for query to BSE */
 bool attach;
 PacketAgentInfo original_dst; /* This was the agent to which the join request
		      * was actually sent to */
 double send_time; /* Time when the packet was sent */
};


struct TreeInfo {

  PacketAgentInfo ag;
  int num_children;
  double dist; /* Distance from self to this agent */
  //  long long_dist;
};

struct sriTreeInfo {

  PacketAgentInfo ag;
  int num_children;
  long dist; /* Distance from self to this agent */
};

struct const_join_response_pkt {

  int accept; /* If this was a join accept or a join reject */
  int layer_id;
  int mbr_count;
  int your_id;
  PacketAgentInfo exp_src; /* This was the agent to which the join request
		      * was actually sent to */
};

struct join_response_pkt {

  bool accept; /* If this was a join accept or a join reject */
  //  int your_id;
  int layer_id;
  int mbr_count;
  int your_id;
  TreeInfo agent_arr[MAX_RET_COUNT];
  PacketAgentInfo exp_src; /* This was the agent to which the join request
		      * was actually sent to */
};

struct const_cluster_refresh_pkt {

  int layer_id;
  int mbr_count;

  /* Info about the higher layer, only carried by the cluster leader */
  int hl_mbr_count;

  int is_root;
  int root_xfer; /* True if the root is being transferred */
  int cluster_remove;
};


struct cluster_refresh_pkt {

  int layer_id;
  int mbr_count;
  TreeInfo agent_arr[MAX_RET_COUNT];

  /* Info about the higher layer, only carried by the cluster leader */
  int hl_mbr_count;
  TreeInfo * hl_agent_arr;

  bool is_root;
  bool root_xfer; /* True if the root is being transferred */
  bool cluster_remove;
};

struct const_cluster_merge_pkt {

  int layer_id;
  int mbr_count;
  //TreeInfo agent_arr[MAX_RET_COUNT];
};

struct cluster_merge_pkt {

  int layer_id;
  int mbr_count;
  TreeInfo agent_arr[MAX_RET_COUNT];
};

struct const_ping_query_pkt {
  
  long src_time;
  int lid;
};

struct ping_query_pkt {

  double src_time;
  int lid;
};

struct const_ping_response_pkt {

  int accept;
  int lid;
  long src_time;
  int mbr_count;
  long dist;
  //TreeInfo * agent_arr;
};

struct ping_response_pkt {

  bool accept;
  int lid;
  double src_time;
  int mbr_count;
  double dist;
  TreeInfo * agent_arr;
};

struct const_data_pkt {
  PacketAgentInfo original_src;
  int seq_no;
  int lid;
  int base_lid;
  int data_len;

  //PRM-added
	int type;
	int hop_count;
	int overlay_hop;
	double orig_time;
	unsigned char bitmap[MAX_BITMAP];
  //PRM
};

//PRM
#define		UP_DIR		0
#define		DOWN_DIR	1

struct data_pkt {

  PacketAgentInfo original_src;
  int seq_no;
  int lid;
  //send for layers above and including base_lid;
  int base_lid;
  int data_len;
  char* payload;

  //PRM-added
	int type;
	int hop_count;
	int overlay_hop;
	double orig_time;
	unsigned char bitmap[MAX_BITMAP];
  //PRM

};

//random_p

#define		SLEE_MAX_NUM	100
#define		NUM_BUDDY_BITMAP	50

struct query_random {
	PacketAgentInfo src_id;
	short lid;
	short direction;
	short hop_count;
	short level;
	PacketAgentInfo rand_id;
	int seq_no;
  double send_time;
	unsigned char bitmap[MAX_BITMAP];
};

struct retransmit_pkt {
	PacketAgentInfo src;

	int seq_no;		//last sequence number
  double send_time;
	unsigned char bitmap[MAX_BITMAP];

};

struct dataack_pkt {

  PacketAgentInfo original_src;
  int seq_no;
  
  double send_time;
};


class AppPacket : public Packet {

public :

  static int m_app_pkt_gen_count;
  static int m_pkt_display_period;

  static int m_jq_cnt;
  static int m_jf_cnt;
  static int m_jr_cnt;
  static int m_cr_cnt;
  static int m_cm_cnt;
  static int m_pq_cnt;
  static int m_pr_cnt;
  static int m_pd_cnt;
  static int m_pda_cnt;
  static int m_pu_cnt;

  AppPacketSubType st;
  union {

    join_query_pkt joinq_p;
    join_forward_pkt joinforward_p;
    join_response_pkt joinresp_p;
    cluster_refresh_pkt clusterrefresh_p;
    cluster_merge_pkt clustermerge_p;
    ping_query_pkt pingq_p;
    ping_response_pkt pingresp_p;
    data_pkt data_p;
    data_pkt dataack_p;

//PRM-added
	query_random random_p;
	retransmit_pkt retransmit_p;
	retransmit_pkt adopt_p;
//PRM-added

  } u;

  AppPacket (AppPacketSubType St);
  ~AppPacket (void);

  static void display_pkt_stats (void);
};

#endif 

