/*
 * File: common-agent.h
 * Author: Suman Banerjee <suman@cs.umd.edu>
 * Date: 15th February, 2002
 * Terms: GPL
 *
 * NICE Application Layer Multicast
 */

#ifndef _COMMON_AGENT_H_
#define _COMMON_AGENT_H_

#include "agent.h"
#include "app-packet.h"
#include "linked-list.h"
#include "timer.h"
#include <arpa/inet.h>

//PRM -magic constant
#define     PROB_RAND_SEND  0.05
#define     QUERY_RANDOM_NODE_TIMEOUT   10.3
#define     BITMAP_SIZE     1024
#define     RXMIT_MULT     8
#define     PROACTIVE_INCREMENT     8
#define		USE_CORRELATION	    0

#define		BLOCK_SIZE		5
#define		FEC_SIZE		0
#define		SHOW_ALL		0

#define     USE_MY_RANDOM_SELECTION     1
#define     DETERMINISTIC_PROACTIVE     1
#define     PROB_CONTINUE   0.8

#define     MAX_ITER        10
#define     FLAT_PROACTIVE  0

#define     FRESH_DATA      0 
#define     RAND_SEND_DATA  1
#define     PROACTIVE_DATA  2

#define     CUT_OFF     5000
#define		LONG_TIME	1000000.0

#define     PROB_AGENT_DOWN         0.001
#define     MAX_LATENCY             15

#define     RANDOM_SEND_LID     -1
#define     PROACTIVE_SEND_LID  -2
#define     ADOPT_LID		  	-3

#define		MISS_BASED_CORR		-1
#define		RECV_BASED_CORR		-2

//PRM

class commonAgent : public Agent {

public :

  //bool handle_timer_now;
  //LinkedList<Timer *, int> expired_timer_list; /* All expired timers */

  void send_pkt_wrapper (AppPacket *ap, int aid, struct sockaddr_in agent_addr);
  void rx_pkt_wrapper (AppPacket *ap);

  void display_detailed_pkt_info (AppPacket *ap);
  AppPacket * reclaim_apppacket(char * recv_buffer, int len);
  int flatten_apppacket(char * payload, AppPacket * ap, int dst_ag);
  void free_app_packet(AppPacket *ap);

	int m_bitmap_size;
/*
  void handle_expired_timers();
  void add_expired_timer(Timer * t);
*/

};


//PRM -added
#define MAX_ID_CACHE	100
#define	RXMIT_CACHE_SIZE	512
#define	RAND_NUM_AT_A_TIME	2

class IdCache {
	int recent_aid;
	//int recent_nid;
	
	//PRM
	struct sockaddr_in recent_addr;//not use
	
	int count;

	int aid[MAX_ID_CACHE];
	//int nid[MAX_ID_CACHE];
	
	struct sockaddr_in addr[MAX_ID_CACHE];
	//unsigned long s_addr[MAX_ID_CACHE];

	int level[MAX_ID_CACHE];
	int recv_score[MAX_ID_CACHE];
	int miss_score[MAX_ID_CACHE];
	int sum_level;
	int sum_recv;
	int sum_miss;

	int found( int ai/*, int ni*/);
	int getId();

public:
	IdCache() { clear(); }
	//recent_nid = recent_aid = -1; count = sum_level = sum_score = 0; }

	int getAgentId(int i);
	//int getNodeId(int i);
	//void getNodeId(int ai, struct sockaddr_in * nid);
	struct sockaddr_in getNodeId(int ai);

	//int addId( int ai, int ni, int lvl, int in_rscore, int in_mscore);
	int addId(int ai, struct sockaddr_in * ni, int lvl, int in_rscore, int in_mscore);

	int addId( AppPacket *, int in_rscore, int in_mscore);

	void clear() { recent_aid = -1; count = sum_level = sum_recv = sum_miss = 0 ; }
	int getSize() { return count; }

	void dump();
};

class RxmitCache {
	char valid[RXMIT_CACHE_SIZE];
	int seq_no[RXMIT_CACHE_SIZE];
	double timeout[RXMIT_CACHE_SIZE];
public:
	RxmitCache() { clear(); }
	void clear();	// { memset ( valid, 0, RXMIT_CACHE_SIZE); }
	int lookup( int seq, double now );
	int insert( int seq, double RTT );
};
//PRM

#endif 
  
