/*
 * File: coop-agent.h
 * Author: Suman Banerjee <suman@cs.umd.edu>
 * Date: 15th February, 2002
 * Terms: GPL
 *
 * NICE Application Layer Multicast
 */

#ifndef _COOP_AGENT_H_
#define _COOP_AGENT_H_

// #include <agent.h>
#include "common-agent.h"
#include "timer.h"
#include "layerinfo.h"
#include "app-packet.h"
//#include "../../lib/rtp/rtp.h"
//#include "nicenode.h"

#define LOG_APP_RP
#define LOG_APP_REFRESH
#undef LOG_APP_REFRESH_OLD
#define LOG_MESH_REFRESH

//Constant changed by PRM
//#define PACKET_CACHE_SIZE	5

#define	RTP_MAX_PACKET_LEN	1500

//PRM
//#define PACKET_CACHE_SIZE	256 
//#define PACKET_CACHE_SIZE	2048 
#define PACKET_CACHE_SIZE	1024 

#define		EXTRA_BYTE	5		//for the hacking code in RTP
#define		HOLD_TIME	(60.0)

#define	NUM_AGENT			8096
//PRM

enum CoopState {
  INIT,
  JOIN,
  ATTACHED,
  STATE_UNDEFINED
};

struct COOPLayerArray {

  LayerInfo * arr[MAX_LAYERS];
};
   
// PRM change2
struct PacketCache {

  AgentInfo ag;
  int seq_no;
  bool valid;
  //PRM
//PRM change2
	int data_len;
	//char payload[RTP_MAX_PACKET_LEN];
	char *payload;
//PRM
	int type;
	double orig_time;
  //PRM
};
//PRM

class coopAgent;

class JoinQueryTimer : public Timer {

public :
  coopAgent *ca;

  JoinQueryTimer (void) : Timer () { ca = NULL;};
  JoinQueryTimer (coopAgent *CA) : Timer () { ca = CA;};
  void EventHandler (bool from_handle_expired_timers);
};

class HigherLayerPingIntervalTimer : public Timer {

public :
  coopAgent *ca;

  HigherLayerPingIntervalTimer (void) : Timer () { ca = NULL;};
  HigherLayerPingIntervalTimer (coopAgent *CA) : Timer () { ca = CA;};
  void EventHandler (bool from_handle_expired_timers);
};

class HigherLayerPingResponseTimer : public Timer {

public :
  coopAgent *ca;

  HigherLayerPingResponseTimer (void) : Timer () { ca = NULL;};
  HigherLayerPingResponseTimer (coopAgent *CA) : Timer () { ca = CA;};
  void EventHandler (bool from_handle_expired_timers);
};

//PRM
class RandomQueryTimer : public Timer {
public:
	coopAgent *ca;
	RandomQueryTimer (void) : Timer () { ca = NULL;};
	RandomQueryTimer (coopAgent *CA) : Timer () { ca = CA;};
	void EventHandler (bool from_handle_expired_timers);
};

//PRM

class coopAgent : public commonAgent {

private :

  CoopState state;

  /* During joins */
  int curr_join_q_lid; /* The layer whose info I am currently seeking */
  int target_join_q_lid; /* This is the layer I want to join */
  /* I will basically get the dist info back for these members. Along
   * with it, I will get the cluster info of the immediate lower layer
   * cluster */
  int m_join_resp_received; /* Used to keep track of how many more
			     * responses are expected */
  COOPLayerArray layers;
  JoinQueryTimer jqt;

  RxmitCache rxmitCache;

  bool m_ping_in_progress;
  int m_ping_lid;
  int m_ping_resps_received;
  HigherLayerPingIntervalTimer m_hlpit; /* Period to do higher layer ping
					   to find better point of attachment */
  HigherLayerPingResponseTimer m_hlprt; /* How long to wait for ping resps */

  int m_data_pkt_seq; /* Used to generate source data packet sequence numbers */

  bool m_use_packet_cache;
  PacketCache m_pkt_cache_array[PACKET_CACHE_SIZE];

public :
  // Statistics stuff
  
  double m_last_change[MAX_LAYERS]; // Last time there was a change to that
  // layer
  
  //PRM
 int m_agent_down;
 int pkt_recv_count;
 int m_sent_proactive;
 int m_pkt_deliv;
 int m_all_recv;
 int m_dist_from_src;
 int m_mature;
 double m_latency;
 double m_recent_up_time; 
 double m_total_up_time; 
 int m_was_dead;
 IdCache idCache;

 private:
 //PRM
 int m_max_seq_recv;
 int m_src_aid;
 int m_src_nid;

 //PRM - the following variables may not be useful (maybe used only for simulator)
 RandomQueryTimer qrt;
 int m_num_all_recv_type[3];
 int m_num_pkt_deliv_type[3];
 int m_duplicate[3][3];
 int pk_duplicate;
 int m_parent;

public :
  
  AgentInfo bse; // The bootstrap entity
 
  coopAgent (void);
  ~coopAgent (void);
  void start (void);
  void stop (void);
  void init (int Id, int Index, int port_no, char * bsehostname,
	     char* talker = NULL, unsigned short dport = 0,unsigned long local_addr = 0); 

  inline void set_bse_agent(int agent_id, struct sockaddr_in agent_addr) { 
    bse.agent_id = agent_id; 
    memcpy(& bse.agent_addr, & agent_addr, sizeof(struct sockaddr_in)); 
  };
  int specific_rx_pkt_handler (Packet *p);  
  void specific_send_data_pkt (char* payload, int data_len);

  void handle_join_query_forward (AppPacket *ap);
  void handle_join_response (AppPacket *ap);
  void handle_cluster_refresh (AppPacket * ap);
  void handle_cluster_merge (AppPacket * ap);
  int handle_data_packet (AppPacket * ap);
  void handle_data_ack_packet (AppPacket * ap);

  void handle_ping_query (AppPacket * ap);
  void handle_ping_response (AppPacket * ap);

  void handle_join_query_timeout (void);

  void handle_cluster_refresh_msg_timeout (int lid);
  void handle_cluster_refresh_check_timeout (int lid);

  void handle_higher_layer_ping_response_timeout (void);
  void handle_higher_layer_ping_interval_timeout (void);

  //PRM
  void queryRandomNode();//added
	int isInternal();
	void KillNode();
	//PRM

	void fillBitmap( unsigned char * bitmap, int size, int aid/*, int nid*/);	

private :

  void init_layer_arr (void);

  bool valid_join_query_forward_packet (int qlid, bool attach);
  bool valid_join_response_packet (AppPacket *ap);
  bool valid_cluster_refresh_packet (AppPacket *ap);
  bool valid_cluster_merge_packet (AppPacket *ap);

  void send_join_query_to_cluster (int cluster_lid);

  void init_join_query (int dst_join_lid);
  void continue_join_query (int this_lid);

  void process_received_join_responses (void);

  void join_to_next_higher_layer (int lid);
  void delete_self_from_all_higher_layers (int lid);
  void delete_self_from_layer (LayerInfo *l);

  void split_cluster_using_two_partition (LayerInfo *l);
  void split_cluster_assuming_2k (LayerInfo * l);
  bool do_merge_to_peer (LayerInfo *l);

  void send_cluster_remove(LayerInfo *l, bool root_xfer);

  void send_root_challenge (LayerInfo *l, LayerAgentInfo *la);
  void do_cluster_merge_root_transfer (LayerInfo *l, AgentInfo *new_root);
  void do_internal_root_transfer (LayerInfo *l, LayerAgentInfo *la);
  void send_all_cluster_refresh_packet (LayerInfo *l, bool set_root_xfer, bool is_root, LayerAgentInfo * ignore_ag);
  void send_specific_cluster_refresh_packet (LayerInfo *l, LayerAgentInfo * la, bool set_root_xfer, bool is_root);

  double get_farthest_dist_of_member_to_agent (LayerInfo *l, LayerAgentInfo * la, bool ignore_self, bool use_only_members_known_to_self);
  LayerAgentInfo * find_farthest_agent_to_me_in_layer (LayerInfo * l);
  LayerAgentInfo * find_closest_agent_in_layer (LayerInfo * this_layer, AgentInfo * ignore_agent, bool match_join_resp);
  void flush_layer (int layer_id);
  void flush_lower_layers (LayerInfo *l);

  LayerAgentInfo * find_new_cluster_leader (LayerInfo *l, bool omit_self);

  int count_dist_known_members_for_agent (LayerAgentInfo *la, bool ignore_self);
  int count_dist_known_members_for_me (LayerInfo *l);

  bool are_agent_members_contained_in_mine (LayerInfo *l, LayerAgentInfo * la);
  bool are_my_members_contained_in_agent (LayerInfo *l, LayerAgentInfo * la);

  LayerAgentInfo * check_all_change_cluster_leader (LayerInfo *l);
  bool check_change_cluster_leader (LayerInfo *l, LayerAgentInfo *la_ag, double & la_dist);

  bool match_cluster_to_root_info (LayerInfo *l);

  void add_extra_members_from_alt_root_packet (AppPacket *ap, LayerInfo *l);

  //PRM changed
  //void forward_data_packet (char* payload, int data_len, int src_aid, struct sockaddr_in src_agent_addr, int src_seqno, int base_lid, int pred_aid, bool is_source);
  
  void forward_data_packet (char* payload, int data_len, int src_aid, struct sockaddr_in src_agent_addr, int src_seqno, int base_lid, int pred_aid, bool is_source, int type, double ot, int overlay_hop);

  //PRM changed
  //int send_data_pkt_to_layer (char* payload, int data_len, LayerInfo * l, int src_aid, int base_lid, struct sockaddr_in src_agent_addr, int src_seqno);

  int send_data_pkt_to_layer (char* payload, int data_len, LayerInfo *l, int src_aid, int base_lid, struct sockaddr_in src_agent_addr, int src_seqno ,int type, double ot, int overlay_hop);

  void send_data_pkt_to(char* payload, int data_len, int src_aid, struct sockaddr_in src_agent_addr, int base_lid, int lid, int src_seqno ,int type, double ot, int overlay_hop, int dst_aid, struct sockaddr_in dst_agent_addr);

  void display_cluster_info (LayerInfo *l, int lid);
  void log_cluster_change_info (int lid);

  bool self_check (LayerAgentInfo * la);

  void cancel_higher_layer_ping_timers (void);
  void set_higher_layer_ping_timer (void);

  LayerInfo * highest_layer_self_present (void);
  void do_ping_higher_layer (LayerInfo *l);
  bool process_ping_responses (void);
  int get_ping_response_count_in_layer (LayerInfo * l);
  bool check_all_ping_responses_received (void);

/* KCR-CHANGE */
  double get_avg_dist_of_member_to_agent (LayerInfo *l, LayerAgentInfo * la, bool ignore_self, bool use_only_members_known_to_self);
  double find_avg_agent_dist_to_me_in_layer (LayerInfo * l);


  // Packet cache stuff
  void init_packet_cache (void);

  //PRM
  //bool add_to_packet_cache (int src_aid, int seq_no);
  //bool add_to_packet_cache (int src_aid, /*int src_nid,*/ int seq_no, int type, double ot);
//PRM change
  bool add_to_packet_cache (int src_aid, /*int src_nid,*/ int seq_no, int type, double ot, int data_len, char* payload);
  //PRM

  /*
  int flatten_apppacket(char * payload, Packet * ap, int dst_ag);
  AppPacket * reclaim_apppacket(char * recv_buffer, int len);
  */

//PRM
  bool lookup_packet_cache (int src_aid, /*int src_nid,*/ int seq_no); //added
//PRM change
  int get_packet_cache (int src_aid, /*int src_nid,*/ int seq_no); //added
//PRM change
  void handle_query_random(AppPacket *ap); //added
  void handle_response_random(AppPacket *ap); //added
  void handle_retransmit_request(AppPacket *ap); //added
	
  // void handle_request_adopt(AppPacket *ap); //skip

  int getRandomAgent(int ni); //added
	
  //getRandomNode(int ai);
  //void getRandomNode(int ai, struct sockaddr_in *); //incomplete
  struct sockaddr_in getRandomNode(int ai);

  LayerAgentInfo * pickAgent(int layer_number, int pred); //added

  int hasAgent( int a_id ); //added
	void requestProactive(AppPacket *ap);
	
	//void sleeDisplay();
	void getLatency();

	inline int getNumLayer() { 
		int i;
		for (i = 0; i < MAX_LAYERS;i++) 
			if ( layers.arr[i] == NULL ) break;
		return (i-1);
	}

	void EventHandler(AgentEvent *);
	
	//int randomSend(int said, int snid, int seq, double ot, int aid, int nid, int type);
	int randomSend(char *payload, int data_len, 
int src_aid, struct sockaddr_in src_agent_addr, int src_seq , double ot, 
		       int type); //added

	int getBitmapScore( AppPacket *ap, int *rscore, int *mscore); //added

	//Orphan -not added
	//double  orphan_clock[MAX_ORPHAN];
	//int orphan_aid[MAX_ORPHAN];
	//int orphan_nid[MAX_ORPHAN];
	//char orphan_valid[MAX_ORPHAN];

	//paramters
	float m_very_late;
	int m_use_prm;
	int m_use_nak;
	float m_prob_rand_send;
	float m_fail_prob;
	//int m_bitmap_size;
	int m_num_rand_send;

	void ReadParam();
public:
	void printLayer0();
//PRM

};
#endif 
  
