/*
 * $Revision: 1.1 $
 * $Id: genutils.h,v 1.1 2002/07/02 19:32:22 rbraud Exp $
 */
#ifndef _GENUTILS_H_
#define _GENUTILS_H_

#include <sys/time.h>
#include <unistd.h>
#include "general.h"

/* A list of general utility functions */

/*
 * To bail out of the process after printing "s" on stdout. The exit
 * value is -1.
 */
void panic (char *s);

/*
 * Also bails out of the process, but only if condition "c" is
 * true. It prints out file(f), line number (l) and function name (fn)
 * as output before exiting with (-1).
 */
void panic_if_cond (int c, char *f, int l, char *fn);
#define panic_if(c) panic_if_cond((c), __FILE__, __LINE__, #c)

/*
 * A conditional printer that prints only iff the boolean flag "c" is
 * true. The function takes in variable arguments just like the standard 
 * printf routine.
 */
void c_printf (boolean c, ...);
/*
 * Same as c_printf, but adds a prefix "Warning: " to the input string 
 * before printing out
 */
void c_warning (boolean c, ...);
/* void c_kill (boolean c, ...); Not implemented...do it later */

/*
 * A generic routine that allocates a buffer of size equal to the file 
 * with name "fname" in the char **s and reads the file into the
 * buffer. It returns the length of the contents read in bytes.
 */
long alloc_read (char **s, char *fname);

#endif
