/*
 * File: timer.h
 * Author: Suman Banerjee <suman@cs.umd.edu>
 * Date: 15th February, 2002
 * Terms: GPL
 *
 * NICE Application Layer Multicast
 */

#ifndef _TIMER_H_
#define _TIMER_H_

#include "o_timeout.h"
#include "linked-list.h"

//extern bool glob_handling_expired_timers;
//extern LinkedList<Timer *, int> expired_timer_list; /* All expired timers */

//void (* wrapper_handler)();

enum TimerType {
  NODE_TIMER_PKT_SEND,
  NODE_TIMER_POLL
};

class Timer {

   private :
     static int timer_id_gen;
     static int timer_instance_gen;
     int id;
     int instance; /* A non-decreasing instance identifier */
     bool is_set;
     o_timeoutHandle handle;
     //     void (* wrapper_handler)();
     

   public :
     void *pos; // The pos returned on insertion into the scheduler

     Timer(void);
     virtual ~Timer (void);
     inline int get_id() {return id;};
     inline int get_instance() {return instance;};
     inline void * get_pos() {return pos;};
     void Handler (bool from_handle_expired_timers); /* The generic handler */
     virtual void EventHandler (bool from_handle_expired_timers) = 0; /* Timer specific handler */

     void SetTimer (double rel_time);
     void CancelTimer (void);

};

struct TimerInfo {
  Timer * t;
  int instance;
  TimerInfo (Timer * T, int I) {
    t = T;
    instance = I;
  };
};

TimerInfo * add_to_timer_instance_list (Timer *t);
bool verify_and_remove_timer_instance (int a);
void handle_expired_timers (void);
void add_expired_timer (Timer * t);

void enable_handle_timers(void);
void disable_handle_timers(void);

#endif 

