#ifndef UTIL_DEBUG_H
#define UTIL_DEBUG_H

#include <stdio.h>
#include <errno.h>
#include <stdarg.h>
#include <string.h>

#include "util.h"


/* extern char *sys_errlist[]; */


#define UTIL_CRITICAL    1000
#define UTIL_ERROR   500
#define UTIL_WARNING     200
#define UTIL_INFORMATION 100

#define UTIL_ABORT  5858
#define UTIL_IGNORE 5859
#define UTIL_RETURN 5860

extern FILE *utilDebugFile;
extern int utilDebugLevel;

/* 
   AssertTrue(bool,FailureString,Action) will execute the statement bool
   and, if false, will print FailureString along with the filename, line
   number and statment that indicated the failure.

   After printing this information, if Action equals UTIL_ABORT or UTIL_IGNORE
   then the failure causes program termination or continuation (respectively).
*/
#define UTIL_AssertTrue(bool,FailureString,Action,rtnVal)       \
    {errno = 0;                         \
     if (!(bool))                           \
     {                              \
        fprintf(utilDebugFile,"*********************\n");       \
        fprintf(utilDebugFile,                  \
        "%s at %s:%d.\n",FailureString,__FILE__,__LINE__);  \
        fprintf(utilDebugFile,"   (this was false: %s)\n",#bool);   \
        fprintf(utilDebugFile,"   (errno (%d) was: %s)\n",  \
                errno, strerror(errno));              \
        fprintf(utilDebugFile,"*********************\n");       \
        if (utilDebugFile != stderr)                \
        {                               \
          fprintf(utilDebugFile,"*********************\n"); \
          fprintf(utilDebugFile,                    \
        "%s at %s:%d.\n",FailureString,__FILE__,__LINE__);  \
          fprintf(utilDebugFile,"   (this was false: %s)\n",#bool);\
          fprintf(utilDebugFile,"   (errno (%d) was: %s)\n",    \
            errno, strerror(errno));              \
          fprintf(utilDebugFile,"*********************\n"); \
        }                               \
        switch(Action)                      \
        {                               \
        case UTIL_ABORT: exit(1);               \
        case UTIL_IGNORE: break;                \
        case UTIL_RETURN: return rtnVal;            \
        default: fprintf(stderr,"Unhandled case\n");exit(1);    \
        }                               \
      }                             \
    }

/*
   InfoMessage is a printf-like funtion that prints a debugging message
   if the current debugging level is less-than priorityLabel
*/
void UTIL_InfoMessage(int priorityLabel,char *string,...);

/* 
   UTIL_SetDebugFile() redirects all further debugging output to the file fp 
*/
int UTIL_SetDebugFile(FILE *fp);

/*
   UTIL_SetDebugLevel() sets the current debugging level to newLevel
*/
int UTIL_SetDebugLevel(int newLevel);

int UTIL_ParseArguments(char **argv,int *argc);




/* Some prototypes that Sun has left-out... morons! */
            extern int fprintf(FILE *stream, const char *format, ...);
            extern int fscanf(FILE *stream, const char *format, ...);
            extern int printf(const char *format, ...);
            extern int scanf(const char *format, ...);
            extern int sscanf(const char *s, const char *format, ...);
            extern int vfprintf(FILE *stream, const char *format, va_list arg);
            extern int vprintf(const char *format, va_list arg);
	    extern int fclose(FILE *stream);
	    extern int strncasecmp(const char *s1, const char *s2, size_t n);
	    extern int strcasecmp(const char *s1, const char *s2);




#endif
