/*
 * File: bse-agent.cc
 * Author: Suman Banerjee <suman@cs.umd.edu>
 * Date: 15th February, 2002
 * Terms: GPL
 *
 * NICE Application Layer Multicast
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>

//#include "nicenode.h"
//#include <scheduler.h>
#include "app-packet.h"
#include "bse-agent.h"
#include "setpartition.h"
#include "common.h"
#include "layerinfo.h"
#include "o_timeout.h"
#include "talker_utils.h"

//added-PRM
#include "common-agent.h"
#include "coop-agent.h"
// #define CLOCK 0.00

#define LOG_BSE_START
#define LOG_BSE_STOP

#define LOG_BSE_SEND
#define LOG_BSE_RECV
#define LOG_BSE_JUNK
#define LOG_BSE_PARTITION

#define LOG_BSE_CLUSTER_CHANGE

#define LOG_DATA_PACKET_SEND
#define LOG_DATA_PACKET_RECV

#define MAX_RECV_BYTES 10000
#define UDP_RECV_AGENT_PORT 5000

void bseAgent::init (int Id, int Index, char* drawer, unsigned short dport) {

  char hostname[512];
  struct hostent he;
  
  //PRM
  m_data_pkt_seq =0;
	m_bitmap_size = MAX_BITMAP;
  //PRM

  Agent::init(Id,Index, drawer, dport);
  memset( & udp_recv_agent_addr, 0, sizeof(struct sockaddr_in));
  memset(&drawer, 0, sizeof(struct sockaddr_in));
  udp_recv_agent_addr.sin_family = AF_INET;
   udp_recv_agent_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  gethostname(hostname, 512);
  memcpy(& he, gethostbyname(hostname), sizeof(he));
  //udp_recv_agent_addr.sin_addr.s_addr = inet_addr("192.168.4.5");
  //udp_recv_agent_addr.sin_addr.s_addr = inet_addr("192.168.11.2");
  //udp_recv_agent_addr.sin_addr.s_addr = * ((unsigned long *) he.h_addr_list[0]);
  //  udp_recv_agent_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
  printf("binding to addr %s\n",inet_ntoa(udp_recv_agent_addr.sin_addr));
  udp_recv_agent_addr.sin_port = htons(UDP_RECV_AGENT_PORT);

  t = AGENT_APPLICATION_BSE;
  top_layer = NULL;
  coop_agents_sofar = 0;
  fflush(0);
  return;
}

void bseAgent::start (void) {

  char recv_buffer[MAX_RECV_BYTES];
  struct sockaddr_in sender_addr;
  unsigned int sender_addr_len = sizeof(struct sockaddr_in);

#define MAX_BYTES       128
  char recvfrom_error_str[MAX_BYTES];
  sprintf(recvfrom_error_str,"recvfrom < %d >:", id);

  Agent::start();

  reset_layer_info(0);
  log_cluster_change_info ();
#ifdef LOG_BSE_CLUSTER_CHANGE
  display_top_layer_info();
#endif 

  if ((udp_sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
    printf("socket() failed\n");

  if (bind(udp_sock, (struct sockaddr *) & udp_recv_agent_addr, sizeof(struct sockaddr)) < 0) 
    printf("bind() failed\n");

#ifdef LOG_BSE_START
  printf ("[bse %d ] At %8.4f start\n", id, /* Scheduler::Clock */ my_clock());
#endif 

  //handle_timer_now = false;
  disable_handle_timers();

  for (;;) {
    int recvmsgsize;
    //handling expired timers here....

    handle_expired_timers();    
    //handle_timer_now = true;
    enable_handle_timers();


    while ((recvmsgsize = recvfrom(udp_sock, recv_buffer, MAX_RECV_BYTES, 0, (struct sockaddr *) & sender_addr, & sender_addr_len)) < 0)
      perror(recvfrom_error_str);

    //handle_timer_now = false;
    disable_handle_timers();

    //    printf("Handling sender %s , recvmsgsize %d\n",inet_ntoa(sender_addr.sin_addr), recvmsgsize);
    /*
     { int i;
    for (i=0; i < recvmsgsize; i++)
      printf ("%d: %x\n", i, recv_buffer[i]);
    printf ("\n");
    }
    */
    
    assert(sender_addr_len == sizeof(struct sockaddr_in));
    AppPacket *ap = (AppPacket *) reclaim_apppacket(recv_buffer, recvmsgsize);
    //    memcpy(& ((Packet *)ap)->src_agent_addr, & sender_addr, sender_addr_len);
    generic_rx_pkt_handler(ap);
    fflush(stdout);
    free_app_packet(ap);
  }

  return;
}

void bseAgent::stop (void) {

#ifdef LOG_BSE_STOP
  printf ("[bse %d ] At %8.4f stop\n", id, /* Scheduler::Clock */ my_clock());
#endif 

  assert (started == true);
  if (top_layer != NULL)
    delete top_layer;

  Agent::stop();
fflush(0);
  return;
}

int bseAgent::specific_rx_pkt_handler  (Packet *p) {

  if (p->t != PACKET_APP) {

     printf ("[Err] BSE received unknown packet type\n");
fflush(0);
     return -1;
  }

  AppPacket *ap = (AppPacket *)p;

  rx_pkt_wrapper(ap);

  switch (ap->st) {
    
  case JOIN_QUERY :
    
    handle_join_query(ap);
    break;

  case CLUSTER_REFRESH :
    handle_cluster_refresh(ap);
    break;
    
  case CLUSTER_MERGE :
    handle_cluster_merge(ap);
    break;
    
  case JOIN_FORWARD :
    printf ("[Err] BSE received Join-forward, ignored\n");
    break;

  case PING_QUERY :
  case PING_RESPONSE :
    printf ("[Err] BSE received ping, ignored\n");
    break;

  case PACKET_DATA :
     printf("\n\n$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$\n\n");
    return handle_data_packet(ap);

  case PACKET_DATA_ACK :
   
    handle_data_ack_packet(ap);
    break;

//PRM-added
	case QUERY_RANDOM:
	case RESPONSE_RANDOM:
//fprintf(stderr, "BSE received a random packet\n");
		break;
	case RETRANSMIT_REQUEST:
		handle_retransmit_request(ap);
		break;
//PRM

  default :
    printf ("[Err] BSE received unknown packet subtype\n");
    return -1;
  }
fflush(0);
  return 0;
}

void bseAgent::reset_layer_info (int top_lid) {

  if (top_layer != NULL) {
    top_layer->MakeEmpty();
    top_layer->lid = top_lid;
    // delete top_layer;
#ifdef LOG_BSE_CLUSTER_CHANGE
  printf ("[bse %d ] At %8.4f cluster-info : lid %d : deleted\n", id, /* Scheduler::Clock */ my_clock(), top_layer->lid);
#endif 
  }
  else
    top_layer = new LayerInfo(top_lid,this);

#ifdef LOG_BSE_JUNK
  printf ("[bse %d ] At %8.4f reset-layer : lid %d\n", id, /* Scheduler::Clock */ my_clock(), top_lid);
#endif 

  LayerAgentInfo * self = new LayerAgentInfo (id,udp_recv_agent_addr);
  self->dist = 0.0;
  top_layer->AddClusterRoot(self);
  top_layer->me_in_layer = true;
  top_layer->cr_chkt.CancelTimer();
  top_layer->cr_msgt.CancelTimer();
  top_layer->cr_chkt.SetTimer(CONST_CLUSTER_REFRESH_CHECK_TIMEOUT);
  top_layer->cr_msgt.SetTimer(CONST_CLUSTER_REFRESH_MSG_TIMEOUT);
fflush(0);
  return;
}

void bseAgent::handle_join_query (AppPacket *p) {

  assert (p->st == JOIN_QUERY); // || (p->st == JOIN_FORWARD) );
  /* Join forward not accepted right now. To do so, changes needed:
     See BSE-JF change in coop-agent.cc
  */
  int qlid;
  AgentInfo src_ag;
  AgentInfo original_dst;
  double src_time;

  if (p->st == JOIN_QUERY) {
    qlid = p->u.joinq_p.q_lid;
    src_ag.agent_id = p->src_agent;
    memcpy(& src_ag.agent_addr, &p->src_agent_addr, sizeof(struct sockaddr_in));
    /*
    if (src_ag.agent_id == UNDEFINED_AGENT_ID) {
      src_ag.agent_id = ++ coop_agents_sofar;
    } 
    memcpy (& src_ag.agent_addr, &p->u.joinq_p.reply_addr, sizeof(struct sockaddr_in));
    */
    original_dst.agent_id = id;
    memcpy (& original_dst.agent_addr, & udp_recv_agent_addr, sizeof(struct sockaddr_in));
    src_time = p->u.joinq_p.send_time;
  }
  else {
    qlid = p->u.joinforward_p.q_lid;
    src_ag.agent_id = p->u.joinforward_p.src_ag.agent_id;
    memcpy (& src_ag.agent_addr, & p->u.joinforward_p.src_ag.agent_addr, sizeof(struct sockaddr_in));
    original_dst.agent_id = p->u.joinforward_p.original_dst.agent_id;
    memcpy (& original_dst.agent_addr , & p->u.joinforward_p.original_dst.agent_addr, sizeof (struct sockaddr_in));
    src_time = p->u.joinforward_p.send_time;
  }
    
#ifdef LOG_BSE_RECV
  printf ("[bse %d ] At %8.4f recvd-join-query : < [ %d ] > lid %d ( top lid %d )\n", id, /* Scheduler::Clock */ my_clock(), src_ag.agent_id, qlid, top_layer->lid);
#endif 

  if (top_layer->lid < qlid) {
    // Someone claims to need to join to a higher layer than I know of
    reset_layer_info(qlid);
  }

  bool do_attach = false;
  LayerAgentInfo * new_member = NULL;
  if (top_layer->lid == qlid) {
    do_attach = true;

    new_member = new LayerAgentInfo(src_ag.agent_id, src_ag.agent_addr);

    bool is_new_member = top_layer->AddClusterMember(new_member);
    if (is_new_member == false) {
      delete new_member;
      new_member = top_layer->FindClusterMember(src_ag.agent_id);
      new_member->refresh = true;
    }
    else {
      log_cluster_change_info ();
#ifdef LOG_BSE_CLUSTER_CHANGE
      display_top_layer_info();
#endif 
    }
    //    new_member->dist = /* Scheduler::Clock */ my_clock() - src_time;
    new_member->dist = MAX_RTT;
    struct sockaddr_in their_addr; 
    memcpy(& their_addr, & new_member->ag.agent_addr, sizeof(struct sockaddr_in));
    their_addr.sin_port = htons(DIST_EST_PORT);
    new_member->dist = estimate_rtt_wrapper(their_addr);
    //if (is_new_member == false)
    //  return;
  }
  /* Create response packet */
  AppPacket *resp_p = new AppPacket(JOIN_RESPONSE);
  resp_p->u.joinresp_p.your_id = src_ag.agent_id; //send new id
  put_layer_cluster_info_into_packet(top_layer,resp_p);
  resp_p->u.joinresp_p.accept = true;
  resp_p->u.joinresp_p.exp_src.agent_id = original_dst.agent_id;
  memcpy ( & resp_p->u.joinresp_p.exp_src.agent_addr, & original_dst.agent_addr, sizeof(struct sockaddr_in));

#ifdef LOG_BSE_RECV
  printf ("[bse %d ] At %8.4f recvd-join-query (sending_resp) : < [ %d ] > lid %d ( top lid %d )\n", id, /* Scheduler::Clock */ my_clock(), src_ag.agent_id, qlid, top_layer->lid);
	printf("to id: %d addr: %s port: %d\n", src_ag.agent_id, inet_ntoa(src_ag.agent_addr.sin_addr), ntohs(src_ag.agent_addr.sin_port ));
#endif 

  send_pkt_wrapper(resp_p,src_ag.agent_id, src_ag.agent_addr);

  if (do_attach == true) {
#ifdef LOG_BSE_SEND
    printf ("\n[bse %d ] At %8.4f : new-member < [ %d ] >\n", id, /* Scheduler::Clock */ my_clock(), new_member->ag.agent_id);
#endif 
    top_layer->cr_msgt.CancelTimer();
    send_top_layer_cluster_refresh(new_member);
    top_layer->cr_msgt.SetTimer(CONST_CLUSTER_REFRESH_MSG_TIMEOUT);
  }
fflush(0);
  return;
}

// Accept the merge only if this is the only member in the top layer
// or if there is no member
void bseAgent::handle_cluster_merge (AppPacket * ap) {

  if ( top_layer->lid != ap->u.clustermerge_p.layer_id )
    return;

  if ( (top_layer->ag_list.GetSize() == 1) ||
       ( (top_layer->ag_list.GetSize() == 2) && (top_layer->ag_list.Locate(ap->src_agent) != NULL) ) ) {
#ifdef LOG_BSE_RECV
    printf ("[bse %d ] At %8.4f recvd-valid-merge : < [ %d ] > lid %d\n", id, /* Scheduler::Clock */ my_clock(), ap->src_agent, ap->u.clustermerge_p.layer_id);
#endif 
    reset_layer_info(ap->u.clustermerge_p.layer_id);
    top_layer->AddExtraMembersFromAltRootPacket(ap);

    log_cluster_change_info ();
#ifdef LOG_BSE_CLUSTER_CHANGE
    display_top_layer_info();
#endif 
  }
  fflush(0);
  return;
}

void bseAgent::handle_cluster_refresh (AppPacket *ap) {

  if (top_layer->lid != ap->u.clusterrefresh_p.layer_id)
    return;
  void * pos = top_layer->ag_list.Locate(ap->src_agent);
  if (pos == NULL)
    return;

#ifdef LOG_BSE_RECV
  printf ("[bse %d ] At %8.4f recvd-valid-refresh : < [ %d ] > lid %d ", id, /* Scheduler::Clock */ my_clock(), ap->src_agent, top_layer->lid);
#endif 

  LayerAgentInfo *la_ag = top_layer->ag_list.GetAt(pos);
  if (ap->u.clusterrefresh_p.cluster_remove == true) {
#ifdef LOG_BSE_RECV
    printf ("*remove*\n");
#endif 
    bool check = top_layer->DeleteClusterMember(la_ag);
    assert (check == true);
    delete la_ag;
    log_cluster_change_info ();
#ifdef LOG_BSE_CLUSTER_CHANGE
    display_top_layer_info();
#endif 
  }
  else {
#ifdef LOG_BSE_RECV
    printf ("\n");
#endif 
    la_ag->refresh_agent(ap);
  }
fflush(0);
  return;
}

void bseAgent::handle_data_ack_packet (AppPacket *ap) {

  assert (ap->st == PACKET_DATA_ACK);
#ifdef LOG_DATA_PACKET_RECV
  printf ("[bse %d ] At %8.4f recv ack-pkt : < [ %d ] > seq %d from < [ %d ] >\n", id, /* Scheduler::Clock */ my_clock(), ap->u.dataack_p.original_src.agent_id, ap->u.dataack_p.seq_no, ap->src_agent);
#endif 
  return;
}

int bseAgent::handle_data_packet (AppPacket *ap) {

  assert (ap->st == PACKET_DATA);
  assert (ap->u.data_p.original_src.agent_id == ap->src_agent);

#ifdef LOG_DATA_PACKET_RECV
  printf ("[bse %d ] At %8.4f recv data-pkt : < [ %d ] > seq %d from < [ %d ] >\n", id, /* Scheduler::Clock */ my_clock(), ap->u.data_p.original_src.agent_id, ap->u.data_p.seq_no, ap->src_agent);
#endif 
  assert (top_layer != NULL);

/* Send an ack back */
  AppPacket *new_ack_p = new AppPacket(PACKET_DATA_ACK);
  
// some changes
	
	new_ack_p->src_agent = this->id;
	new_ack_p->src_agent_addr = this->udp_recv_agent_addr;
	
	//

  new_ack_p->u.dataack_p.seq_no = ap->u.data_p.seq_no;
  new_ack_p->u.dataack_p.original_src.agent_id = ap->u.data_p.original_src.agent_id;
  memcpy((void*)&(new_ack_p->u.dataack_p.original_src.agent_addr), (void*)&(ap->u.data_p.original_src.agent_addr), sizeof(struct sockaddr_in));
  send_pkt_wrapper(new_ack_p, ap->src_agent, ap->src_agent_addr);

  fflush(0);
  if (top_layer->lid < ap->u.data_p.base_lid)
    return 0;

  for (void * pos = top_layer->ag_list.GetHeadPosition();
       pos != NULL;
       top_layer->ag_list.GetNext(pos) ) {

    LayerAgentInfo * this_la = top_layer->ag_list.GetAt(pos);

    if (this_la->ag.agent_id == ap->u.data_p.original_src.agent_id)
       continue;

    if (self_check(this_la) == false) {

       AppPacket *new_p = new AppPacket(PACKET_DATA);
       new_p->u.data_p.original_src.agent_id = ap->u.data_p.original_src.agent_id;
       memcpy ( & new_p->u.data_p.original_src.agent_addr, & ap->u.data_p.original_src.agent_addr, sizeof(struct sockaddr_in));
       new_p->u.data_p.seq_no = ap->u.data_p.seq_no;
       new_p->u.data_p.lid = top_layer->lid;
       new_p->u.data_p.base_lid = ap->u.data_p.base_lid;
#ifdef LOG_DATA_PACKET_SEND
       // KCR-CHANGE
       //printf ("[bse %d ] At %8.4f send data-pkt : < [ %d ] > seq %d : to < [ %d ] > lid %d dist_est %f \n", id, /* Scheduler::Clock */ my_clock(), ap->u.data_p.original_src.agent_id, ap->u.data_p.seq_no, this_la->ag.agent_id, top_layer->lid, this_la->dist);
#endif 
       //send_pkt_wrapper(new_p,this_la->ag.agent_id, this_la->ag.agent_addr);
       delete new_p;
    }
  }
  fflush(0);
  return 0;
}
 

void bseAgent::handle_cluster_refresh_msg_timeout (void) {

  assert (top_layer != NULL);
#ifdef LOG_BSE_SEND
  printf ("\n[slee-bse %d ] At %8.4f : refresh-timeout: lid %d\n", id, /* Scheduler::Clock */ my_clock(), top_layer->lid);
#endif 

  send_top_layer_cluster_refresh(NULL);
  top_layer->cr_msgt.SetTimer(CONST_CLUSTER_REFRESH_MSG_TIMEOUT);
  fflush(0);
  return;
}

void bseAgent::send_top_layer_cluster_refresh (LayerAgentInfo * ignore_agent) {

  assert (top_layer != NULL);

#ifdef LOG_BSE_SEND
  printf ("\n[bse %d ] At %8.4f (b-scr) : send-refresh : lid %d\n", id, /* Scheduler::Clock */ my_clock(), top_layer->lid);
#endif 

  for (void * pos = top_layer->ag_list.GetHeadPosition();
       pos != NULL;
       top_layer->ag_list.GetNext(pos) ) {

    LayerAgentInfo * la = top_layer->ag_list.GetAt(pos);
    if (ignore_agent != NULL) {
      if (ignore_agent->ag.agent_id == la->ag.agent_id)
         continue;
    }

    if (self_check(la) == false) {

#ifdef LOG_BSE_SEND
  printf ("[bse %d ] . . (b-scr) : < [ %d ] >\n", id, la->ag.agent_id);
#endif 

      AppPacket *ap = new AppPacket (CLUSTER_REFRESH);
      put_layer_cluster_info_into_packet(top_layer,ap);
      ap->u.clusterrefresh_p.root_xfer = false;
      ap->u.clusterrefresh_p.is_root = true;
      send_pkt_wrapper(ap,la->ag.agent_id,la->ag.agent_addr);
    }
  }
#ifdef LOG_BSE_SEND
  printf ("\n");
#endif 
  fflush(0);
  return;
}


void bseAgent::handle_cluster_refresh_check_timeout (void) {

  bool change = false;

  assert (top_layer != NULL);
#ifdef LOG_BSE_JUNK
      printf ("[bse %d] At %8.4f (b-rc) : refresh-chk : lid %d\n", id, /* Scheduler::Clock */ my_clock(), top_layer->lid);
#endif 
fflush(0);
  for (void * pos = top_layer->ag_list.GetHeadPosition();
       pos != NULL;
       ) {

    LayerAgentInfo * la_ag = top_layer->ag_list.GetAt(pos);
    void * old_pos = pos;
    top_layer->ag_list.GetNext(pos);
    if (self_check(la_ag) == true)
      continue;
    if (la_ag->refresh == false) {
      /* Member is lost */
      top_layer->ag_list.RemoveAt(old_pos);
#ifdef LOG_BSE_JUNK
      printf ("[bse %d ] . . (b-rc) : < [ %d ] > *lost*\n", id, la_ag->ag.agent_id);
#endif 

      delete la_ag;
      change = true;
    }
    else {
      la_ag->refresh = false; // Resetting the refresh flag
#ifdef LOG_BSE_JUNK
      printf ("[bse %d ] . . (b-rc) : < [ %d ] >\n", id, la_ag->ag.agent_id);
#endif 
    }
  }
#ifdef LOG_BSE_JUNK
  printf ("\n");
#endif 

  assert (top_layer->ag_list.GetSize() >= 1);
  if (top_layer->ag_list.GetSize() == 1) { // I am the only one
    if (top_layer->lid > 0) {
#ifdef LOG_BSE_CLUSTER_CHANGE
  printf ("[bse %d ] At %8.4f cluster-info : lid %d : deleted\n", id,/* Scheduler::Clock */ my_clock(), top_layer->lid);
#endif 
      top_layer->lid --;
#ifdef LOG_BSE_JUNK
      printf ("[bse %d ] At %8.4f refresh-chk layer-decrement to lid %d\n", id, /* Scheduler::Clock */ my_clock(), top_layer->lid);
#endif 
      change = true;
    }
  }

  if (top_layer->ag_list.GetSize() >= (UPPER_3K+1)) {
    split_top_layer_using_two_partition();
    change = true;
  }

  if (change == true) {
    log_cluster_change_info ();
#ifdef LOG_BSE_CLUSTER_CHANGE
    display_top_layer_info();
#endif 
  }

  top_layer->cr_chkt.SetTimer(CONST_CLUSTER_REFRESH_CHECK_TIMEOUT);
  fflush(0);
  return;
}

void bseAgent::split_top_layer_using_two_partition (void) {

#ifdef LOG_BSE_JUNK
  printf ("[bse %d ] At %8.4f cluster-split : lid %d\n", id, /* Scheduler::Clock */ my_clock(), top_layer->lid);
#endif 

  LayerAgentInfo ** ag_arr = top_layer->CreateLayerAgentInfoArray();
  assert (ag_arr != NULL);
  double * self_included_cost = create_cost_matrix(id,ag_arr,top_layer->ag_list.GetSize());
  // This function already updates the agent_arr by deleting this agent
  double * cost = delete_agent_from_cost_matrix(self_included_cost,ag_arr,top_layer->ag_list.GetSize(),id);
  free(self_included_cost);

  void * pos = top_layer->ag_list.Locate(id);
  assert (pos != NULL);
  LayerAgentInfo * self = top_layer->ag_list.GetAt(pos);
  bool check = top_layer->DeleteClusterMember(self);
  assert (check == true);

  int root1_index, root2_index;
  int * index_set1;
  int * index_set2;
  int set1_size, set2_size;

  // Partition the set of members
  int lower_size = top_layer->ag_list.GetSize() / 2;
  two_partition(top_layer->ag_list.GetSize(),cost,lower_size,index_set1,set1_size,root1_index,index_set2,set2_size,root2_index);
  free(cost);

#ifdef LOG_BSE_PARTITION
  printf ("[bse %d ] . . cluster-split : Partition 1 : ", id);
  for (int i = 0; i < set1_size; i++)
    printf ("%d ", ag_arr[index_set1[i]]->ag.agent_id);
  printf("\n");
  printf ("[bse %d ] . . cluster-split : Partition 2 : ", id);
  for (int i = 0; i < set2_size; i++)
    printf ("%d ", ag_arr[index_set2[i]]->ag.agent_id);
  printf("\n");
#endif 
  
  LayerInfo * tmp_layer1 = new LayerInfo(top_layer->lid,NULL);
  LayerInfo * tmp_layer2 = new LayerInfo(top_layer->lid,NULL);
  for (int i = 0; i < set1_size; i++) {
    top_layer->DeleteClusterMember(ag_arr[index_set1[i]]);
    bool success = tmp_layer1->AddClusterMember(ag_arr[index_set1[i]]);
    assert (success == true);
  }
  tmp_layer1->root = ag_arr[root1_index];

  for (int i = 0; i < set2_size; i++) {
    top_layer->DeleteClusterMember(ag_arr[index_set2[i]]);
    bool success = tmp_layer2->AddClusterMember(ag_arr[index_set2[i]]);
    assert (success == true);
  }
  tmp_layer2->root = ag_arr[root2_index];

  free(index_set1);
  free(index_set2);

#ifdef LOG_BSE_PARTITION
  printf ("[bse %d ] At %8.4f cluster-split : root1 < [ %d ] > count %d     root2 < [ %d ] > count %d\n", id, /* Scheduler::Clock */ my_clock(), ag_arr[root1_index]->ag.agent_id, set1_size, ag_arr[root2_index]->ag.agent_id, set2_size);
#endif 

  free(ag_arr);

  assert (top_layer->ag_list.GetSize() == 0);

  // Transfer control of the other cluster
  send_cluster_remove(tmp_layer1);
  send_cluster_remove(tmp_layer2);

  LayerAgentInfo * new_root1 = new LayerAgentInfo(tmp_layer1->root->ag.agent_id,tmp_layer1->root->ag.agent_addr);
  LayerAgentInfo * new_root2 = new LayerAgentInfo(tmp_layer2->root->ag.agent_id,tmp_layer2->root->ag.agent_addr);

  // Cleanup ...
  delete tmp_layer1;
  delete tmp_layer2;

  assert (top_layer->lid + 1 < MAX_LAYERS);
#ifdef LOG_BSE_CLUSTER_CHANGE
  printf ("[bse %d ] At %8.4f cluster-info : lid %d : deleted\n", id, /* Scheduler::Clock */ my_clock(), top_layer->lid);
#endif 
#ifdef LOG_BSE_JUNK
  printf ("[bse %d ] At %8.4f layer-increment : lid %d\n", id, /* Scheduler::Clock */ my_clock(), top_layer->lid+1);
#endif 

  // Setup the next higher layer
  top_layer->lid ++;
  top_layer->AddClusterRoot(self);
  top_layer->me_in_layer = true;

  top_layer->AddClusterMember(new_root1);
  top_layer->AddClusterMember(new_root2);
  fflush(0);
  return;
}

void bseAgent::send_cluster_remove (LayerInfo * l) {

#ifdef LOG_BSE_SEND
      printf ("[bse %d ] At %8.4f (b-scx) : self-cluster-remove : lid %d : new-root < [ %d ] >\n", id, /* Scheduler::Clock */ my_clock(), top_layer->lid, l->root->ag.agent_id);
#endif 
  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       l->ag_list.GetNext(pos) ) {

    LayerAgentInfo * la = l->ag_list.GetAt(pos);
    assert (self_check(la) == false);
    AppPacket *ap = new AppPacket (CLUSTER_REFRESH);
    put_cluster_remove_info_into_packet(l,ap,true,NULL,id,udp_recv_agent_addr);

#ifdef LOG_BSE_SEND
    printf ("[bse %d ] . . (b-scx) : < [ %d ] >\n", id, la->ag.agent_id);
#endif 

    send_pkt_wrapper(ap,la->ag.agent_id,la->ag.agent_addr);
  }
#ifdef LOG_BSE_SEND
  printf ("\n");
#endif 
  fflush(0);
  return;
}

void bseAgent::log_cluster_change_info (void) {

  m_last_change = /* Scheduler::Clock */ my_clock();
  return;
}

void bseAgent::display_top_layer_info (void) {

  assert (top_layer != NULL);

  printf ("[bse %d ] At %8.4f cluster-info : lid %d : ldr %s %d : count %d : ", id, /* Scheduler::Clock */ my_clock(), top_layer->lid, (self_check(top_layer->root) == true) ? "self" : "other", top_layer->root->ag.agent_id, top_layer->ag_list.GetSize() );

#ifdef LOG_CLUSTER_INFO_DETAILED
  struct_pkt_reset();
  struct_pkt_set_size(top_layer->ag_list.GetSize());
  struct_pkt_set_lid(top_layer->lid);
  struct_pkt_set_ldr(top_layer->root->ag.agent_id);
  pkt_info.my_id = 0;
  for (void * pos = top_layer->ag_list.GetHeadPosition();
       pos != NULL;
       top_layer->ag_list.GetNext(pos) ) {
    int tid = (top_layer->ag_list.GetAt(pos))->ag.agent_id;
    printf (" %d", tid);
    struct_pkt_add_member(tid);
  }
  send_struct_pkt();
#endif 

  printf ("\n");
  fflush(0);
  return;
}

bool bseAgent::self_check (LayerAgentInfo * la) {

  if (la->ag.agent_id == id) 
    return true;
  else
    return false;
}

//PRM
int bseAgent::send_data_pkt_to_layer (LayerInfo * l, int src_seqno) {

  int count_pkts = 0;
  double now = /*Scheduler::Clock();*/my_clock();

	m_clock_cache[src_seqno % CLOCK_CACHE_SIZE] = now;

	//fprintf(stderr, "%d\n", src_seqno);


	//do some bitmap stuff here
	int i, base;
	char bitmap[MAX_BITMAP];
	for ( i = 0, base = src_seqno - m_bitmap_size; i < m_bitmap_size; i++ ) {
		if ( base+i < 0 )
			continue;
		bitmap[i] = 1;	//i am the source
	}

  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       l->ag_list.GetNext(pos) ) {

    LayerAgentInfo * this_la = l->ag_list.GetAt(pos);
    if ( this_la->ag.agent_id != this->id ) {
       AppPacket *ap = new AppPacket(PACKET_DATA);
       ap->u.data_p.original_src.agent_id = this->id;
       //ap->u.data_p.original_src.node_id = this->n->id;
       memcpy(&ap->u.data_p.original_src.agent_addr, &this->udp_recv_agent_addr, sizeof(struct sockaddr_in));
  
       ap->u.data_p.seq_no = src_seqno;
       ap->u.data_p.lid = l->lid;
		ap->u.data_p.type = FRESH_DATA;
		ap->u.data_p.orig_time = now;
		memcpy( ap->u.data_p.bitmap, bitmap, m_bitmap_size * sizeof (char) );

#ifdef LOG_DATA_PACKET_SEND
		/*     printf ("[coop %d %d ] At %8.4f send data-pkt : < [ %d %d ] > seq %d : to < [ %d %d ] > lid %d\n", id, n->id, Scheduler::Clock(), src_aid, src_nid, src_seqno, this_la->ag.agent_id, this_la->ag.node_id, l->lid);*/
#endif //LOG_DATA_PACKET_SEND
		//send_pkt_wrapper(ap,this_la->ag.agent_id, this_la->ag.node_id);
		send_pkt_wrapper(ap,this_la->ag.agent_id,this_la->ag.agent_addr);
       count_pkts ++;
   }
 }

  return count_pkts;
}


//PRM
void bseAgent::specific_send_data_pkt() {
  assert (started == true);

/*
#ifdef LOG_DATA_PACKET_INIT
  printf ("[coop %d %d ] At %8.4f source data-pkt : seq %d\n", id, n->id, Scheduler::Clock(), m_data_pkt_seq);
#endif LOG_DATA_PACKET_INIT
*/

  /* THe following should be used as stats only (NOT SURE)
	if ( m_data_pkt_seq < CUT_OFF )
		up_per_pkt[m_data_pkt_seq] = UpCount;
  */
	if ( top_layer != NULL )
		send_data_pkt_to_layer ( top_layer, m_data_pkt_seq++);

	/*
	if ( top_layer && m_data_pkt_seq%1000 == 1 )
		fprintf(stderr, "bse: first packet top_layer = %d\n", top_layer->lid );
	*/

  return;
}

void bseAgent::handle_retransmit_request( AppPacket *ap ) {
	int seq = ap->u.retransmit_p.seq_no;
	AppPacket *nap;
	PacketAgentInfo *pai = &ap->u.retransmit_p.src;
	unsigned char *bitmap = ap->u.retransmit_p.bitmap;
	int cseq;

	if ( pai->agent_id != id )
		return;

	for ( int i = 0 ; i < m_bitmap_size; i++ ) {
		if ( (cseq = seq+i-m_bitmap_size ) < 0 )
			continue;
		if ( bitmap[i] && cseq < m_data_pkt_seq && cseq >= m_data_pkt_seq - CLOCK_CACHE_SIZE) {
       		nap = new AppPacket(PACKET_DATA);

		// some changes
	
		nap->src_agent = this->id;
		nap->src_agent_addr = this->udp_recv_agent_addr;
	
		//

       		nap->u.data_p.original_src = *pai;
       		nap->u.data_p.seq_no = cseq;
       		nap->u.data_p.lid = PROACTIVE_SEND_LID;
			nap->u.data_p.type = PROACTIVE_DATA;
	   		nap->u.data_p.orig_time = m_clock_cache[cseq % CLOCK_CACHE_SIZE];

			//send_pkt_wrapper(nap, ap->src_agent, ap->src );
			send_pkt_wrapper(nap, ap->src_agent, ap->src_agent_addr );
     	}
	}
}

//PRM-added







