/*
 * File: common-agent.cc
 * Author: Suman Banerjee <suman@cs.umd.edu>
 * Date: 15th February, 2002
 * Terms: GPL
 *
 * NICE Application Layer Multicast
 */

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <sys/types.h>
#include <netinet/in.h>

//#include "nicenode.h"
#include "common-agent.h"
#include "app-packet.h"
//#include <scheduler.h>

#define PRECISION_DOUBLE	1000.0

//#define CLOCK 0.00
//#define LOG_DETAILED_PKT_SEND
//#define LOG_DETAILED_PKT_RCV

#define MAX_PAYLOAD 10000
#define HOST_INFO_SIZE	6

int	SLEE_DEBUG = 0;
//#define slee_printf(...) { if (0) printf( __VA_ARGS__ ); }

void commonAgent::send_pkt_wrapper (AppPacket *ap, int aid, struct sockaddr_in agent_addr) {

  char payload[MAX_PAYLOAD];
#ifdef LOG_DETAILED_PKT_SEND
     printf ("[ %d ] spd %f [ %d ] tp %d", id, /* Scheduler::my_clock() */ my_clock(), aid, ap->st);
     display_detailed_pkt_info(ap);
     fflush(stdout);
#endif
    //printf("app-packet type is %d\n",ap->st);

  int len = flatten_apppacket(payload, ap, aid);

    //printf("flattened app-packet type is %d\n",ap->st);
    //display_detailed_pkt_info(ap);
  
    //display_detailed_pkt_info(app);

/*
     if (ap->st == JOIN_QUERY)
       printf(" slee: JOIN-QUERY to agent %d IP-addr %s port %d\n",aid, inet_ntoa(agent_addr.sin_addr), ntohs(((sockaddr_in) agent_addr).sin_port));
     if (ap->st == JOIN_RESPONSE)
       printf(" slee: JOIN-RESPONSE to agent %d IP-addr %s port %d\n",aid, inet_ntoa(agent_addr.sin_addr), ntohs(((sockaddr_in) agent_addr).sin_port));
*/
       
     //printf ("[ %d ] spd2 %f [ %d ] tp %d", ap->src_agent, /* Scheduler::my_clock() */ my_clock(), aid, ap->st);
  
  send_pkt(payload,len,aid,agent_addr);
    //slee_printf(" slee: sent payload to agent %d port %d\n",aid, ntohs(((sockaddr_in) agent_addr).sin_port));
  free_app_packet(ap);
  return;
}

void commonAgent::rx_pkt_wrapper (AppPacket *ap) {
#ifdef LOG_DETAILED_PKT_RCV
   //slee_printf ("[ %d ] rpd %f [ %d ] tp %d", id, /* Scheduler::my_clock() */ my_clock(), ap->src_agent, ap->st);
   display_detailed_pkt_info(ap);
    fflush(stdout);

#endif
  return;
}

void commonAgent::display_detailed_pkt_info (AppPacket *ap) {
  switch (ap->st) {

  case JOIN_QUERY:
    printf (" sz 5 + 0\n");
    break;

  case JOIN_FORWARD:
    printf (" sz %d + 0\n", 2 * HOST_INFO_SIZE +  9);
    break;

  case PING_QUERY:
    printf (" sz 5 + 0\n");
    break;

  case JOIN_RESPONSE:
    printf (" sz %d + %d\n", HOST_INFO_SIZE + 7, HOST_INFO_SIZE *
ap->u.joinresp_p.mbr_count);
    break;

  case CLUSTER_REFRESH:
    printf (" sz 13 + %d\n", HOST_INFO_SIZE *
(ap->u.clusterrefresh_p.mbr_count+ap->u.clusterrefresh_p.hl_mbr_count));
    break;

  case CLUSTER_MERGE:
    printf (" sz 5 + %d\n", HOST_INFO_SIZE * ap->u.clustermerge_p.mbr_count);
    break;

  case PING_RESPONSE:
    printf (" sz 9 + %d\n", HOST_INFO_SIZE * ap->u.pingresp_p.mbr_count);
    break;

  case PACKET_DATA:
    printf(" sz %d + 0\n", 2 + HOST_INFO_SIZE);
    break;

  case PACKET_DATA_ACK:
    printf(" sz %d + 0\n", 2 + HOST_INFO_SIZE);
    break;

//PRM
	case QUERY_RANDOM:
		printf("  query_random\n");
		break;
	case RESPONSE_RANDOM:
		printf("  response_random\n");
		break;
	case RETRANSMIT_REQUEST:
		printf("  retransmit_request\n");
		break;
//PRM

  default:

    assert(0);
  }
  fflush(stdout);

  return;
}


AppPacket * commonAgent::reclaim_apppacket(char * recv_buffer, int len) {

  enum AppPacketSubType st; 
  memcpy(& st, recv_buffer, sizeof(enum AppPacketSubType));
  st = (enum AppPacketSubType) ntohl(st);
  AppPacket *ap = new AppPacket(st); 
  int newlen = sizeof(enum AppPacketSubType);

  //  printf("within reclaim: type %d\n",ap->st);
  memcpy(& ap->src_agent, recv_buffer + newlen, sizeof(int));
  ap->src_agent = (int) ntohl(ap->src_agent);
  newlen += sizeof(int);

  memcpy(&ap->src_agent_addr, recv_buffer + newlen, sizeof(struct sockaddr_in));
  newlen += sizeof(struct sockaddr_in);

  memcpy(& ap->dst_agent, recv_buffer + newlen, sizeof(int));
  ap->dst_agent = (int) ntohl(ap->dst_agent);
  newlen += sizeof(int);
    
  switch (ap->st) {
  case (JOIN_QUERY) : {
    
    struct const_join_query_pkt * data = (struct const_join_query_pkt *) ((char *)recv_buffer + newlen);
    
    
        //memcpy(& ap->u.joinq_p, data, sizeof(struct join_query_pkt));
	ap->u.joinq_p.q_lid = (int) ntohl(data->q_lid);
    //    ap->u.joinq_p.q_lid = (int) ntohl(ap->u.joinq_p.q_lid);
    ap->u.joinq_p.attach = (bool) (ntohl(data->attach));
        //slee_printf("send time is %f\n",data->send_time);

    //ap->u.joinq_p.send_time = ((double) (long) ntohl(data->send_time) )/ PRECISION_DOUBLE;
       memcpy ((char *)&ap->u.joinq_p.send_time, (char *)&data->send_time, sizeof(double));


        //slee_printf("send time is %f\n",ap->u.joinq_p.send_time);
    memcpy(& ap->u.joinq_p.reply_addr, & data->reply_addr, sizeof(struct sockaddr_in));
    newlen += sizeof(struct const_join_query_pkt);
  }
    break;

  case (JOIN_FORWARD) : {
   
    struct const_join_forward_pkt * data = (struct const_join_forward_pkt *) ((char *)recv_buffer + newlen);
    ap->u.joinforward_p.src_ag.agent_id = ntohl(data->src_ag.agent_id);
    memcpy(&ap->u.joinforward_p.src_ag.agent_addr, &data->src_ag.agent_addr, sizeof(struct sockaddr_in));
    ap->u.joinforward_p.q_lid = (int) ntohl(data->q_lid);
    ap->u.joinforward_p.attach = (bool) (ntohl(data->attach));
    ap->u.joinforward_p.original_dst.agent_id = ntohl(data->original_dst.agent_id);
    memcpy(&ap->u.joinforward_p.original_dst.agent_addr, &data->original_dst.agent_addr, sizeof(struct sockaddr_in));
    //ap->u.joinforward_p.send_time = ((double) (long) ntohl(data->send_time)) / PRECISION_DOUBLE;
        memcpy(& ap->u.joinforward_p.send_time, & data->send_time, sizeof(double));
    newlen += sizeof(struct const_join_forward_pkt);
  }
    break;

  case (JOIN_RESPONSE) : {
    
    struct const_join_response_pkt * data = (struct const_join_response_pkt *) ((char *) recv_buffer + newlen);
    //    memcpy(& ap->u.joinresp_p, data, sizeof(struct const_join_response_pkt));
    ap->u.joinresp_p.accept = (bool)(ntohl(data->accept));
    ap->u.joinresp_p.layer_id = (int) ntohl(data->layer_id);
    ap->u.joinresp_p.mbr_count = (int) ntohl(data->mbr_count);
    ap->u.joinresp_p.your_id = (int) ntohl(data->your_id);
    /*    ap->u.joinresp_p.layer_id = (int) ntohl(ap->u.joinresp_p.layer_id);
    ap->u.joinresp_p.mbr_count = (int) ntohl(ap->u.joinresp_p.mbr_count);
    ap->u.joinresp_p.your_id = (int) ntohl(ap->u.joinresp_p.your_id);
    */
    ap->u.joinresp_p.exp_src.agent_id = ntohl(data->exp_src.agent_id);
    memcpy(&ap->u.joinresp_p.exp_src.agent_addr, &data->exp_src.agent_addr, sizeof(struct sockaddr_in));
    newlen += sizeof(struct const_join_response_pkt);

    //    memcpy(ap->u.joinresp_p.agent_arr, (recv_buffer + newlen), (ap->u.joinresp_p.mbr_count)*sizeof(struct TreeInfo));
    for (int i = 0; i < ap->u.joinresp_p.mbr_count; i++) {
      struct sriTreeInfo * tempinfo = (struct sriTreeInfo *) ((char *) recv_buffer + newlen);
      ap->u.joinresp_p.agent_arr[i].num_children = (int) ntohl(tempinfo[i].num_children);
      ap->u.joinresp_p.agent_arr[i].ag.agent_id = (int) ntohl(tempinfo[i].ag.agent_id);
      memcpy(& ap->u.joinresp_p.agent_arr[i].ag.agent_addr, & tempinfo[i].ag.agent_addr, sizeof(struct sockaddr_in));
      ap->u.joinresp_p.agent_arr[i].dist = ((double) (long) ntohl(tempinfo[i].dist)) / PRECISION_DOUBLE;
    }
    newlen += (ap->u.joinresp_p.mbr_count)*sizeof(struct sriTreeInfo);

    /*
    printf("reclaiming debug:\n");
    printf("mbr_count %d mbrs : ",ap->u.joinresp_p.mbr_count);
    for (int i = 0; i < ap->u.joinresp_p.mbr_count; i++) {
      printf("%d ",ap->u.joinresp_p.agent_arr[i].ag.agent_id);
    }
    printf("\n");
    */
  }
    break;

  case (CLUSTER_REFRESH) : {
    
    struct const_cluster_refresh_pkt * data = (struct const_cluster_refresh_pkt *) ((char *)recv_buffer + newlen);
    ap->u.clusterrefresh_p.layer_id = (int) ntohl(data->layer_id);
    ap->u.clusterrefresh_p.mbr_count = (int) ntohl(data->mbr_count);
    ap->u.clusterrefresh_p.hl_mbr_count = (int) ntohl(data->hl_mbr_count); 
    ap->u.clusterrefresh_p.is_root = (bool) ntohl(data->is_root);
    ap->u.clusterrefresh_p.root_xfer = (bool) ntohl(data->root_xfer);
    ap->u.clusterrefresh_p.cluster_remove = (bool) ntohl(data->cluster_remove);
    newlen += sizeof(struct const_cluster_refresh_pkt);

    //    memcpy(ap->u.clusterrefresh_p.agent_arr, recv_buffer + newlen, (ap->u.clusterrefresh_p.mbr_count)*sizeof(struct TreeInfo));
    for (int i = 0; i < ap->u.clusterrefresh_p.mbr_count; i++) {
      struct sriTreeInfo * tempinfo = (struct sriTreeInfo *) ((char *) recv_buffer + newlen);
      ap->u.clusterrefresh_p.agent_arr[i].num_children = (int) ntohl(tempinfo[i].num_children);
      ap->u.clusterrefresh_p.agent_arr[i].ag.agent_id = (int) ntohl(tempinfo[i].ag.agent_id);
      memcpy(& ap->u.clusterrefresh_p.agent_arr[i].ag.agent_addr, & tempinfo[i].ag.agent_addr, sizeof(struct sockaddr_in));
      ap->u.clusterrefresh_p.agent_arr[i].dist = (double) (long) ntohl(tempinfo[i].dist);
    }
    newlen += (ap->u.clusterrefresh_p.mbr_count)*sizeof(struct sriTreeInfo);

    ap->u.clusterrefresh_p.hl_agent_arr = (struct TreeInfo *) malloc((ap->u.clusterrefresh_p.hl_mbr_count)*sizeof(struct TreeInfo));
    //    memcpy(ap->u.clusterrefresh_p.hl_agent_arr, recv_buffer + newlen, (ap->u.clusterrefresh_p.hl_mbr_count)*sizeof(struct TreeInfo));
    for (int i = 0; i < ap->u.clusterrefresh_p.hl_mbr_count; i++) {
      struct sriTreeInfo * tempinfo = (struct sriTreeInfo *) ((char *) recv_buffer + newlen);
      ap->u.clusterrefresh_p.hl_agent_arr[i].num_children = (int) ntohl(tempinfo[i].num_children);
      ap->u.clusterrefresh_p.hl_agent_arr[i].ag.agent_id = (int) ntohl(tempinfo[i].ag.agent_id);
      memcpy(& ap->u.clusterrefresh_p.hl_agent_arr[i].ag.agent_addr, & tempinfo[i].ag.agent_addr, sizeof(struct sockaddr_in));
      ap->u.clusterrefresh_p.hl_agent_arr[i].dist = (double) (long) ntohl(tempinfo[i].dist);
    }
    newlen += (ap->u.clusterrefresh_p.hl_mbr_count)*sizeof(struct sriTreeInfo);

  }
    break; 

  case (CLUSTER_MERGE) : {

    struct const_cluster_merge_pkt * data = (struct const_cluster_merge_pkt *) ((char *)recv_buffer + newlen);
    ap->u.clustermerge_p.layer_id = (int) ntohl(data->layer_id); 
    ap->u.clustermerge_p.mbr_count = (int) ntohl(data->mbr_count); 
    newlen += sizeof(struct const_cluster_merge_pkt);

    memcpy(ap->u.clustermerge_p.agent_arr, recv_buffer + newlen, (ap->u.clustermerge_p.mbr_count)*sizeof(struct TreeInfo));
    for (int i = 0; i < ap->u.clustermerge_p.mbr_count; i++) {
      struct sriTreeInfo * tempinfo = (struct sriTreeInfo *) ((char *) recv_buffer + newlen);
      ap->u.clustermerge_p.agent_arr[i].num_children = (int) ntohl(tempinfo[i].num_children);
      ap->u.clustermerge_p.agent_arr[i].ag.agent_id = (int) ntohl(tempinfo[i].ag.agent_id);
      memcpy(& ap->u.clustermerge_p.agent_arr[i].ag.agent_addr, & tempinfo[i].ag.agent_addr, sizeof(struct sockaddr_in));
      ap->u.clustermerge_p.agent_arr[i].dist = (double) (long) ntohl(tempinfo[i].dist);
    }
    newlen += (ap->u.clustermerge_p.mbr_count)*sizeof(struct sriTreeInfo);

  }
    break; 
    
  case (PING_QUERY) : {

    struct const_ping_query_pkt * data = (struct const_ping_query_pkt *) ((char *)recv_buffer + newlen);
    ap->u.pingq_p.src_time = ((double) (long) ntohl(data->src_time)) / PRECISION_DOUBLE;
    //    memcpy(& ap->u.pingq_p.src_time, & data->src_time, sizeof(double)); 
    ap->u.pingq_p.lid = (int) ntohl(data->lid);
    newlen += sizeof(struct ping_query_pkt);
  }
    break; 

  case (PING_RESPONSE) : {

    struct const_ping_response_pkt * data = (struct const_ping_response_pkt *) ((char *)recv_buffer + newlen);
    ap->u.pingresp_p.accept = (bool) ntohl(data->accept); 
    ap->u.pingresp_p.lid = (int) ntohl(data->lid);
    ap->u.pingresp_p.src_time = ( (double) (long) ntohl(data->src_time) ) / PRECISION_DOUBLE;
    //    memcpy(& ap->u.pingresp_p.src_time, & data->src_time, sizeof(double));
    ap->u.pingresp_p.mbr_count = (int) ntohl(data->mbr_count);
    ap->u.pingresp_p.dist = ( (double) (long) ntohl(data->dist) ) / PRECISION_DOUBLE;
    newlen += sizeof(struct join_query_pkt);

    ap->u.pingresp_p.agent_arr = (struct TreeInfo *) malloc((ap->u.pingresp_p.mbr_count)*sizeof(struct TreeInfo));
    memcpy(ap->u.pingresp_p.agent_arr, recv_buffer + newlen, (ap->u.pingresp_p.mbr_count)*sizeof(struct TreeInfo));
    for (int i = 0; i < ap->u.pingresp_p.mbr_count; i++) {
      struct sriTreeInfo * tempinfo = (struct sriTreeInfo *) ((char *) recv_buffer + newlen);
      ap->u.pingresp_p.agent_arr[i].num_children = (int) ntohl(tempinfo[i].num_children);
      ap->u.pingresp_p.agent_arr[i].ag.agent_id = (int) ntohl(tempinfo[i].ag.agent_id);
      memcpy(& ap->u.pingresp_p.agent_arr[i].ag.agent_addr, & tempinfo[i].ag.agent_addr, sizeof(struct sockaddr_in));
      ap->u.pingresp_p.agent_arr[i].dist = (double) (long) ntohl(tempinfo[i].dist);
    }
    newlen += (ap->u.pingresp_p.mbr_count)*sizeof(struct sriTreeInfo);

  }
    break;

  case (PACKET_DATA) : {

    struct const_data_pkt * data = (struct const_data_pkt *) ((char *)recv_buffer + newlen);
    ap->u.data_p.original_src.agent_id = ntohl(data->original_src.agent_id);
    memcpy(&ap->u.data_p.original_src.agent_addr, &data->original_src.agent_addr, sizeof(struct sockaddr_in));
    ap->u.data_p.seq_no = (int) ntohl(data->seq_no); 
    ap->u.data_p.lid = (int) ntohl(data->lid); 
    ap->u.data_p.base_lid = (int) ntohl(data->base_lid);
    ap->u.data_p.data_len = (int) ntohl(data->data_len);
//PRM
    	ap->u.data_p.type = (int) ntohl(data->type); 

//commonAgent::reclaim_apppacket(char * recv_buffer, int len) {

    	ap->u.data_p.hop_count = (int) ntohl(data->hop_count); 
    	ap->u.data_p.overlay_hop = (int) ntohl(data->overlay_hop); 
        memcpy(& ap->u.data_p.orig_time, & data->orig_time, sizeof(double));
	
    newlen += sizeof(struct const_data_pkt);
	newlen -= MAX_BITMAP;
//PRM
    /* copy variable part */

    //koo
   	memcpy(ap->u.data_p.bitmap, data->bitmap, m_bitmap_size*sizeof(char)/8);
	newlen += m_bitmap_size/8;
	
    ap->u.data_p.payload = (char *)malloc(ap->u.data_p.data_len);
    //ap->u.data_p.payload = (recv_buffer+newlen);
    memcpy(ap->u.data_p.payload, recv_buffer + newlen, ap->u.data_p.data_len);
    
    newlen += ap->u.data_p.data_len;
  }
    break; 

  case (PACKET_DATA_ACK) : {
    
    struct dataack_pkt * data = (struct dataack_pkt *) ((char *)recv_buffer + newlen);
    ap->u.dataack_p.original_src.agent_id = ntohl(data->original_src.agent_id);
    memcpy(&ap->u.dataack_p.original_src.agent_addr, &data->original_src.agent_addr, sizeof(struct sockaddr_in));
    ap->u.dataack_p.seq_no = (int) ntohl(data->seq_no); 
    newlen += sizeof(struct dataack_pkt);
    /* copy variable part */
    /* len += sizeof(variable part) */
  
    //Time added
    memcpy(&ap->u.dataack_p.orig_time, & data->send_time, sizeof(double));
    
  }
    break; 

//PRM
	case (QUERY_RANDOM) : {
    		struct query_random * data = (struct query_random *) ((char *)recv_buffer + newlen);
    		ap->u.random_p.src_id.agent_id = ntohl(data->src_id.agent_id);
    		memcpy(&ap->u.random_p.src_id.agent_addr, &data->src_id.agent_addr, sizeof(struct sockaddr_in));
    		ap->u.random_p.lid = (short) ntohl(data->lid); 
    		ap->u.random_p.direction = (short) ntohl(data->direction); 
    		ap->u.random_p.hop_count = (short) ntohl(data->hop_count); 
    		ap->u.random_p.level = (short) ntohl(data->level); 

    		ap->u.random_p.rand_id.agent_id = ntohl(data->rand_id.agent_id);
    		memcpy(&ap->u.random_p.rand_id.agent_addr, &data->rand_id.agent_addr, sizeof(struct sockaddr_in));
    		ap->u.random_p.seq_no = (int) ntohl(data->seq_no); 
    		memcpy(ap->u.random_p.bitmap, data->bitmap, m_bitmap_size*sizeof(char)/8);

    		newlen += sizeof(struct query_random);
			newlen -= MAX_BITMAP;
			newlen += m_bitmap_size/8;

		//Time added
		memcpy(& ap->u.random_p.send_time, & data->send_time, sizeof(double));
	}
		break;

	case (RESPONSE_RANDOM) : {
    		struct query_random * data = (struct query_random *) ((char *)recv_buffer + newlen);
    		ap->u.random_p.src_id.agent_id = ntohl(data->src_id.agent_id);
    		memcpy(&ap->u.random_p.src_id.agent_addr, &data->src_id.agent_addr, sizeof(struct sockaddr_in));
    		ap->u.random_p.lid = (short) ntohl(data->lid); 
    		ap->u.random_p.direction = (short) ntohl(data->direction); 
    		ap->u.random_p.hop_count = (short) ntohl(data->hop_count); 
    		ap->u.random_p.level = (short) ntohl(data->level); 

    		ap->u.random_p.rand_id.agent_id = ntohl(data->rand_id.agent_id);
    		memcpy(&ap->u.random_p.rand_id.agent_addr, &data->rand_id.agent_addr, sizeof(struct sockaddr_in));
    		ap->u.random_p.seq_no = (int) ntohl(data->seq_no); 
    		memcpy(ap->u.random_p.bitmap, data->bitmap, m_bitmap_size*sizeof(char)/8);

    		newlen += sizeof(struct query_random);
			newlen -= MAX_BITMAP;
			newlen += m_bitmap_size/8;

		 //Time added
		memcpy(& ap->u.random_p.send_time, & data->send_time, sizeof(double));
	}
		break;

	case (RETRANSMIT_REQUEST) : {
    		struct retransmit_pkt * data = (struct retransmit_pkt *) ((char *)recv_buffer + newlen);
    		ap->u.retransmit_p.src.agent_id = ntohl(data->src.agent_id);
    		memcpy(&ap->u.retransmit_p.src.agent_addr, &data->src.agent_addr, sizeof(struct sockaddr_in));
    		ap->u.retransmit_p.seq_no = (int) ntohl(data->seq_no); 
//reclaim_packet
    		memcpy(ap->u.retransmit_p.bitmap, data->bitmap, m_bitmap_size*sizeof(char)/8);
    		newlen += sizeof(struct retransmit_pkt);
			newlen -= MAX_BITMAP;
			newlen += m_bitmap_size/8;

		//Time added
		memcpy(& ap->u.retransmit_p.send_time, & data->send_time, sizeof(double));
	}
		break;
//PRM
      
  case (PACKET_UNDEFINED) :
       
    printf("[Err] undefined packet\n");
    free(ap);
    ap = NULL;
    break; 
      
  default : 

    /* printf("undefined packet\n"); */
    free(ap);
    ap = NULL;
    break;

  }
  fflush(stdout);
  return ap;
}

int commonAgent::flatten_apppacket(char * payload, AppPacket * ap, int dst_ag) {

  int len = sizeof(ap->st);
  long int tempint;
  tempint = htonl(ap->st);
  memcpy(payload, & tempint, len);

  tempint = htonl(id);
  memcpy(payload + len, & tempint, sizeof(int));
  len += sizeof(int);
  
  memcpy(payload + len, & udp_recv_agent_addr, sizeof(struct sockaddr_in));
  len += sizeof(struct sockaddr_in);
  
  assert(ntohs(udp_recv_agent_addr.sin_port) == (unsigned short) (id + 5000));

  tempint = htonl(dst_ag);
  memcpy(payload + len, & tempint, sizeof(int));
  len += sizeof(int);

  switch (ap->st) {
  case (JOIN_QUERY) : {
	
    struct const_join_query_pkt * data = (struct const_join_query_pkt *) ((char *)payload + len);
    int temp = htonl(ap->u.joinq_p.q_lid);
    data->q_lid = temp;
    data->attach = htonl((int) ap->u.joinq_p.attach);
    //data->send_time = (long) htonl((long) (ap->u.joinq_p.send_time * PRECISION_DOUBLE));
        memcpy(& data->send_time, & ap->u.joinq_p.send_time, sizeof(data->send_time));
    memcpy(& data->reply_addr, & ap->u.joinq_p.reply_addr, sizeof(struct sockaddr_in));
    len += sizeof(struct const_join_query_pkt);
  }
    break;

  case (JOIN_FORWARD) : {
	
    struct const_join_forward_pkt * data = (struct const_join_forward_pkt *) ((char *)payload + len);
    data->src_ag.agent_id = htonl(ap->u.joinforward_p.src_ag.agent_id);
    memcpy(&data->src_ag.agent_addr, &ap->u.joinforward_p.src_ag.agent_addr, sizeof(struct sockaddr_in));
    assert((unsigned short) (ap->u.joinforward_p.src_ag.agent_id + 5000) == ntohs(ap->u.joinforward_p.src_ag.agent_addr.sin_port));
    int temp = htonl(ap->u.joinforward_p.q_lid);
    data->q_lid = temp;
    /* data->q_lid = (int) htonl(data->q_lid); BIG ERROR */

    data->attach = htonl((int) ap->u.joinforward_p.attach);
    data->original_dst.agent_id = htonl(ap->u.joinforward_p.original_dst.agent_id);
    memcpy(&data->original_dst.agent_addr, &ap->u.joinforward_p.original_dst.agent_addr, sizeof(struct sockaddr_in));
    assert((unsigned short) (5000 + ap->u.joinforward_p.original_dst.agent_id) == ntohs(ap->u.joinforward_p.original_dst.agent_addr.sin_port));
    //data->send_time = (long) htonl((long) (ap->u.joinforward_p.send_time * PRECISION_DOUBLE));
      
    // made a change here

  memcpy(& data->send_time, & ap->u.joinforward_p.send_time, sizeof(double));
    len += sizeof(struct const_join_forward_pkt);
  }
    break;
	
  case (JOIN_RESPONSE) : {
	
    struct const_join_response_pkt * data = (struct const_join_response_pkt *) ((char *)payload + len);
    data->accept = htonl((int) ap->u.joinresp_p.accept);
    int temp = htonl(ap->u.joinresp_p.layer_id);
    data->layer_id = temp;
    temp = htonl(ap->u.joinresp_p.mbr_count);
    data->mbr_count = temp;
    temp = htonl(ap->u.joinresp_p.your_id);
    data->your_id = temp;

    data->exp_src.agent_id = htonl(ap->u.joinresp_p.exp_src.agent_id);
    memcpy(&data->exp_src.agent_addr, &ap->u.joinresp_p.exp_src.agent_addr, sizeof(struct sockaddr_in));
    assert((unsigned short) (ap->u.joinresp_p.exp_src.agent_id + 5000) == ntohs(ap->u.joinresp_p.exp_src.agent_addr.sin_port));
    len += sizeof(struct const_join_response_pkt);

    //    memcpy(payload + len, ap->u.joinresp_p.agent_arr, (ap->u.joinresp_p.mbr_count)*sizeof(struct TreeInfo));
    for (int i = 0; i < ap->u.joinresp_p.mbr_count; i++) {
      struct sriTreeInfo * tempinfo = (struct sriTreeInfo *)(payload + len);
      tempinfo[i].ag.agent_id = (int) htonl(ap->u.joinresp_p.agent_arr[i].ag.agent_id);
      assert((unsigned short) (5000 + ap->u.joinresp_p.agent_arr[i].ag.agent_id) == ntohs(ap->u.joinresp_p.agent_arr[i].ag.agent_addr.sin_port));
      memcpy( & tempinfo[i].ag.agent_addr, & ap->u.joinresp_p.agent_arr[i].ag.agent_addr, sizeof(struct sockaddr_in));
      tempinfo[i].num_children = (int) htonl(ap->u.joinresp_p.agent_arr[i].num_children);
      tempinfo[i].dist = (long) htonl((long) (ap->u.joinresp_p.agent_arr[i].dist * PRECISION_DOUBLE));

//        double temp_double;
//        memcpy(& temp_double, & tempinfo[i].dist, sizeof(double));
//        tempinfo[i].long_dist = htonl((long) temp_double);
    }
    len += (ap->u.joinresp_p.mbr_count)*sizeof(struct sriTreeInfo);

    /*
    printf("flattening debug:\n");
    printf("mbr_count %d mbrs : ",ap->u.joinresp_p.mbr_count);
    for (int i = 0; i < ap->u.joinresp_p.mbr_count; i++) {
      printf("%d ",ap->u.joinresp_p.agent_arr[i].ag.agent_id);
    }
    printf("\n");
    */

  }
    break;

  case (CLUSTER_REFRESH) : {

    struct const_cluster_refresh_pkt * data = (struct const_cluster_refresh_pkt *) ((char *)payload + len);
    data->layer_id = (int) htonl(ap->u.clusterrefresh_p.layer_id);
    data->mbr_count = (int) htonl(ap->u.clusterrefresh_p.mbr_count);
    data->hl_mbr_count = (int) htonl(ap->u.clusterrefresh_p.hl_mbr_count);
    data->is_root = htonl((int)ap->u.clusterrefresh_p.is_root);
    data->root_xfer = htonl((int)ap->u.clusterrefresh_p.root_xfer);
    data->cluster_remove = htonl((int)ap->u.clusterrefresh_p.cluster_remove);
    len += sizeof(struct const_cluster_refresh_pkt);

    //    memcpy(payload + len, ap->u.clusterrefresh_p.agent_arr, (ap->u.clusterrefresh_p.mbr_count)*sizeof(struct TreeInfo));
    for (int i = 0; i < ap->u.clusterrefresh_p.mbr_count; i++) { 
      struct sriTreeInfo * tempinfo = (struct sriTreeInfo *)(payload + len);
      tempinfo[i].ag.agent_id = (int) htonl(ap->u.clusterrefresh_p.agent_arr[i].ag.agent_id);
      assert((unsigned short) (5000 + ap->u.clusterrefresh_p.agent_arr[i].ag.agent_id) == ntohs(ap->u.clusterrefresh_p.agent_arr[i].ag.agent_addr.sin_port));
      memcpy( & tempinfo[i].ag.agent_addr, & ap->u.clusterrefresh_p.agent_arr[i].ag.agent_addr, sizeof(struct sockaddr_in));
      tempinfo[i].num_children = (int) htonl(ap->u.clusterrefresh_p.agent_arr[i].num_children);
      tempinfo[i].dist = (long) htonl((long) ap->u.clusterrefresh_p.agent_arr[i].dist);
    }
    len += (ap->u.clusterrefresh_p.mbr_count)*sizeof(struct sriTreeInfo);

    //    memcpy(payload + len, ap->u.clusterrefresh_p.hl_agent_arr, (ap->u.clusterrefresh_p.hl_mbr_count)*sizeof(struct TreeInfo));
    for (int i = 0; i < ap->u.clusterrefresh_p.hl_mbr_count; i++) { 
      struct sriTreeInfo * tempinfo = (struct sriTreeInfo *)(payload + len);
      tempinfo[i].ag.agent_id = (int) htonl(ap->u.clusterrefresh_p.hl_agent_arr[i].ag.agent_id);
      assert((unsigned short) (5000 + ap->u.clusterrefresh_p.hl_agent_arr[i].ag.agent_id) == ntohs(ap->u.clusterrefresh_p.hl_agent_arr[i].ag.agent_addr.sin_port));
      memcpy( & tempinfo[i].ag.agent_addr, & ap->u.clusterrefresh_p.hl_agent_arr[i].ag.agent_addr, sizeof(struct sockaddr_in));
      tempinfo[i].num_children = (int) htonl(ap->u.clusterrefresh_p.hl_agent_arr[i].num_children);
      tempinfo[i].dist = (long) htonl((long) ap->u.clusterrefresh_p.hl_agent_arr[i].dist);
    }
    len += (ap->u.clusterrefresh_p.hl_mbr_count)*sizeof(struct sriTreeInfo);
  }
    break; 

  case (CLUSTER_MERGE) : {
    
    struct const_cluster_merge_pkt * data = (struct const_cluster_merge_pkt *) ((char *)payload + len);
    data->layer_id = (int) htonl(ap->u.clustermerge_p.layer_id);
    data->mbr_count = (int) htonl(ap->u.clustermerge_p.mbr_count);
    len += sizeof(struct const_cluster_merge_pkt);

    //    memcpy(payload + len, ap->u.clustermerge_p.agent_arr, (ap->u.clustermerge_p.mbr_count)*sizeof(struct TreeInfo));
    for (int i = 0; i < ap->u.clustermerge_p.mbr_count; i++) { 
      struct sriTreeInfo * tempinfo = (struct sriTreeInfo *)(payload + len);
      tempinfo[i].ag.agent_id = (int) htonl(ap->u.clustermerge_p.agent_arr[i].ag.agent_id);
      assert((unsigned short) (5000 + ap->u.clustermerge_p.agent_arr[i].ag.agent_id) == ntohs(ap->u.clustermerge_p.agent_arr[i].ag.agent_addr.sin_port));
      memcpy( & tempinfo[i].ag.agent_addr, & ap->u.clustermerge_p.agent_arr[i].ag.agent_addr, sizeof(struct sockaddr_in));
      tempinfo[i].num_children = (int) htonl(ap->u.clustermerge_p.agent_arr[i].num_children);
      tempinfo[i].dist = (long) htonl((long) ap->u.clustermerge_p.agent_arr[i].dist);
    }
    len += (ap->u.clustermerge_p.mbr_count)*sizeof(struct sriTreeInfo);

  }
    break; 
    
  case (PING_QUERY) : {

    struct const_ping_query_pkt * data = (struct const_ping_query_pkt *) ((char *)payload + len);
    data->src_time = (long) htonl((long) (ap->u.pingq_p.src_time * PRECISION_DOUBLE));
    //    memcpy(& data->src_time, & ap->u.pingq_p.src_time, sizeof(double));
    data->lid = (int) htonl(ap->u.pingq_p.lid);
    len += sizeof(struct ping_query_pkt);
  }
    break; 

  case (PING_RESPONSE) : {

    struct const_ping_response_pkt * data = (struct const_ping_response_pkt *) ((char *)payload + len);
    data->accept = htonl((int)ap->u.pingresp_p.accept);
    data->lid = (int) htonl(ap->u.pingresp_p.lid);
    data->src_time = (long) htonl((long) (ap->u.pingresp_p.src_time * PRECISION_DOUBLE) );
    //    memcpy( & data->src_time, & ap->u.pingresp_p.src_time, sizeof(double));
    data->mbr_count = (int) htonl(ap->u.pingresp_p.mbr_count);
    data->dist = (long) htonl((long) (ap->u.pingresp_p.dist * PRECISION_DOUBLE) );
    //    memcpy(& data->dist, & ap->u.pingresp_p.dist, sizeof(double));
    len += sizeof(struct join_query_pkt);

    //    memcpy(payload + len, ap->u.pingresp_p.agent_arr, (ap->u.pingresp_p.mbr_count)*sizeof(struct TreeInfo));
    for (int i = 0; i < ap->u.pingresp_p.mbr_count; i++) { 
      struct sriTreeInfo * tempinfo = (struct sriTreeInfo *)(payload + len);
      tempinfo[i].ag.agent_id = (int) htonl(ap->u.pingresp_p.agent_arr[i].ag.agent_id);
      assert((unsigned short) (5000 + ap->u.pingresp_p.agent_arr[i].ag.agent_id) == ntohs(ap->u.pingresp_p.agent_arr[i].ag.agent_addr.sin_port));
      memcpy( & tempinfo[i].ag.agent_addr, & ap->u.pingresp_p.agent_arr[i].ag.agent_addr, sizeof(struct sockaddr_in));
      tempinfo[i].num_children = (int) htonl(ap->u.pingresp_p.agent_arr[i].num_children);
      tempinfo[i].dist = (long) htonl((long) ap->u.pingresp_p.agent_arr[i].dist);
    }
    len += (ap->u.pingresp_p.mbr_count)*sizeof(struct sriTreeInfo);

  }
    break;

  case (PACKET_DATA) : {

    struct const_data_pkt * data = (struct const_data_pkt *) (payload + len);
    data->original_src.agent_id = htonl(ap->u.data_p.original_src.agent_id);
    memcpy(&data->original_src.agent_addr, &ap->u.data_p.original_src.agent_addr, sizeof(struct sockaddr_in));
    assert((unsigned short) (5000 + ap->u.data_p.original_src.agent_id) == ntohs(ap->u.data_p.original_src.agent_addr.sin_port));
    data->seq_no = htonl(ap->u.data_p.seq_no);
    data->lid = htonl(ap->u.data_p.lid);
    data->base_lid = htonl(ap->u.data_p.base_lid);
    data->data_len = htonl(ap->u.data_p.data_len);

     data->type = htonl(ap->u.data_p.type);

//commonAgent::flatten_apppacket(char * payload, AppPacket * ap, int dst_ag) {

    data->hop_count = htonl(ap->u.data_p.hop_count);
    data->overlay_hop = htonl(ap->u.data_p.overlay_hop);
     //Time added
    memcpy(& data->orig_time, & ap->u.data_p.orig_time, sizeof(double));

    len += sizeof(struct const_data_pkt);
	len -= MAX_BITMAP;

    /* copy variable part */
    memcpy(data->bitmap,ap->u.data_p.bitmap, m_bitmap_size*sizeof(char)/8);
	len += m_bitmap_size/8;
    memcpy(payload + len, ap->u.data_p.payload, ap->u.data_p.data_len);
    len += ap->u.data_p.data_len;

   
  }
    break; 
      
  case (PACKET_DATA_ACK) : {

    struct dataack_pkt * data = (struct dataack_pkt *) ((char *)payload + len);
    data->original_src.agent_id = htonl(ap->u.dataack_p.original_src.agent_id);
    memcpy(&data->original_src.agent_addr, &ap->u.dataack_p.original_src.agent_addr, sizeof(struct sockaddr_in));
    assert((unsigned short) (5000 + ap->u.dataack_p.original_src.agent_id) == ntohs(ap->u.dataack_p.original_src.agent_addr.sin_port));
    data->seq_no = htonl(ap->u.dataack_p.seq_no);
    len += sizeof(struct dataack_pkt);
    /* copy variable part */
    /* len += sizeof(variable part) */

    //Time added
    memcpy(& data->send_time, & ap->u.dataack_p.orig_time, sizeof(double));
  
  }
    break; 

//PRM
	case (QUERY_RANDOM) : {
    		struct query_random * data = (struct query_random *) ((char *)payload + len);
    		data->src_id.agent_id = htonl(ap->u.random_p.src_id.agent_id);
    		memcpy(&data->src_id.agent_addr, &ap->u.random_p.src_id.agent_addr, sizeof(struct sockaddr_in));
    		data->lid = htonl(ap->u.random_p.lid);
    		data->direction = htonl(ap->u.random_p.direction);
    		data->hop_count = htonl(ap->u.random_p.hop_count);
    		data->level = htonl(ap->u.random_p.level);

    		data->rand_id.agent_id = htonl(ap->u.random_p.rand_id.agent_id);
    		memcpy(&data->rand_id.agent_addr, &ap->u.random_p.rand_id.agent_addr, sizeof(struct sockaddr_in));
		
		//Time added
		memcpy(& data->send_time, & ap->u.random_p.send_time, sizeof(double));
  

    		assert((unsigned short) (5000 + ap->u.random_p.src_id.agent_id) == ntohs(ap->u.random_p.src_id.agent_addr.sin_port));
    		data->seq_no = htonl(ap->u.random_p.seq_no);
    		memcpy(data->bitmap, ap->u.random_p.bitmap, m_bitmap_size*sizeof(char)/8);
    		len += sizeof(struct query_random);
			len -= MAX_BITMAP;
			len += m_bitmap_size/8;


	}
	break;
	case (RESPONSE_RANDOM) : {
    		struct query_random * data = (struct query_random *) ((char *)payload + len);
    		data->src_id.agent_id = htonl(ap->u.random_p.src_id.agent_id);
    		memcpy(&data->src_id.agent_addr, &ap->u.random_p.src_id.agent_addr, sizeof(struct sockaddr_in));
    		data->lid = htonl(ap->u.random_p.lid);
    		data->direction = htonl(ap->u.random_p.direction);
    		data->hop_count = htonl(ap->u.random_p.hop_count);
    		data->level = htonl(ap->u.random_p.level);

    		data->rand_id.agent_id = htonl(ap->u.random_p.rand_id.agent_id);
    		memcpy(&data->rand_id.agent_addr, &ap->u.random_p.rand_id.agent_addr, sizeof(struct sockaddr_in));
		
		
		//Time added
		memcpy(& data->send_time, & ap->u.random_p.send_time, sizeof(double));
  
    		assert((unsigned short) (5000 + ap->u.random_p.src_id.agent_id) == ntohs(ap->u.random_p.src_id.agent_addr.sin_port));
    		data->seq_no = htonl(ap->u.random_p.seq_no);
    		memcpy(data->bitmap, ap->u.random_p.bitmap, m_bitmap_size*sizeof(char)/8);
    		len += sizeof(struct query_random);
			len -= MAX_BITMAP;
			len += m_bitmap_size/8;
	}
	break;
	case (RETRANSMIT_REQUEST) : {
    		struct retransmit_pkt * data = (struct retransmit_pkt *) ((char *)payload + len);
    		data->src.agent_id = htonl(ap->u.retransmit_p.src.agent_id);
    		memcpy(&data->src.agent_addr, &ap->u.retransmit_p.src.agent_addr, sizeof(struct sockaddr_in));
    		assert((unsigned short) (5000 + ap->u.retransmit_p.src.agent_id) == ntohs(ap->u.retransmit_p.src.agent_addr.sin_port));
    		data->seq_no = htonl(ap->u.retransmit_p.seq_no);

		
		//Time added
		memcpy(& data->send_time, & ap->u.retransmit_p.send_time, sizeof(double));
  
//commonAgent::flatten_apppacket(char * payload, AppPacket * ap, int dst_ag) {

    		memcpy(data->bitmap, ap->u.retransmit_p.bitmap, m_bitmap_size*sizeof(char)/8);
    		len += sizeof(struct retransmit_pkt);
			len -= MAX_BITMAP;
			len += m_bitmap_size/8;

	}
	break;
//PRM
      
  case (PACKET_UNDEFINED) :
       
    printf("[Err] undefined packet\n");
    break; 
      
  default : 

    /* printf("[Err] undefined packet\n"); */
    break;

  }
  fflush(stdout);
  return len;
}

void commonAgent::free_app_packet(AppPacket *ap) {

  switch (ap->st) {

  case (CLUSTER_REFRESH) : {
    
    if (ap->u.clusterrefresh_p.hl_mbr_count > 0) {
      assert(ap->u.clusterrefresh_p.hl_agent_arr);
      free(ap->u.clusterrefresh_p.hl_agent_arr);
      ap->u.clusterrefresh_p.hl_agent_arr = NULL;
    }
    break; 
  }

  case (PING_RESPONSE) : {

    if (ap->u.pingresp_p.mbr_count > 0) {
      assert(ap->u.pingresp_p.agent_arr);
      free(ap->u.pingresp_p.agent_arr);
      ap->u.pingresp_p.agent_arr = NULL;
    }
  }
    break;

  case (PACKET_UNDEFINED) :
       
    printf("[Err] undefined packet\n");
    break; 

  case (PACKET_DATA) : {
    assert(ap->u.data_p.payload);
    //free(ap->u.data_p.payload);
  }
    break;

  default : 

    /* printf("undefined packet\n"); */
    break;
  }
  assert(ap);
  if(ap->st != PACKET_DATA) {
    delete ap;
    ap = NULL;
  }
  fflush(stdout);
  return;
}

/*
void commonAgent::handle_expired_timers() {
  
  int call_unmask_timer;

  call_unmask_timer = mask_timer();
  glob_handling_expired_timers = true;
  for (void *pos = expired_timer_list.GetHeadPosition();
       pos != NULL;
       expired_timer_list.GetNext(pos) ) {
    Timer *t = expired_timer_list.GetAt(pos);
    if (verify_and_remove_timer_instance(t->get_instance()) == true)
      t->Handler(true);
  }

  expired_timer_list.RemoveAll();
  glob_handling_expired_timers = false;
  if(call_unmask_timer == CALL_UNMASK_TIMER)
    unmask_timer();

  fflush(stdout);
  return;
}

void commonAgent::add_expired_timer(Timer * t) {

  expired_timer_list.Add(t, t->get_id());
  return;
}
*/



//PRM

int IdCache::getAgentId( int ni ) {
	if ( count <= 0 ) {
printf("count is 0 in getAgentId %d\n", count);
		return -1;
	}

	int index = -1;
/*
	if ( USE_CORRELATION ) {
		if ( ni == MISS_BASED_CORR && sum_miss > 0 ) {
			int rnum = rand() % sum_miss;
			for ( int i = 0 ; i < count; i++ ) {
				rnum -= miss_score[i];
				if ( rnum < 0 ) {
					index = i;
					break;
				}
			}
		}
		if ( ni == RECV_BASED_CORR && sum_recv > 0 ) {
			int rnum = rand() % sum_recv;
			for ( int i  = 0 ; i  < count; i ++ ) {
				rnum -= recv_score[i];
				if ( rnum < 0 ) {
					index = i;
					break;
				}
			}
		}
	}
*/

	if ( index < 0 ) {
	 	index = rand() % count;
	}
	recent_aid = aid[index];
	
	return recent_aid;
}


//int IdCache::getNodeId( int ai ) {
struct sockaddr_in IdCache::getNodeId(int ai){

  //	if ( ai >= 0 && recent_aid == ai )
  //		return recent_nid;
 //PRM cache
	//unsigned long agent_addr;
	struct sockaddr_in naddr;
  
  //nid = NULL;

	if ( ai >= 0 )
		for ( int i = 0; i < count; i++ ) {
			if ( aid[i] == ai ) {
				recent_aid = aid[i];
				//recent_nid = nid[i];				
				//return nid[i];

				//naddr = (addr[i]);
//slee_printf("getrand %d node addr %s\n", aid[i], inet_ntoa(addr[i].sin_addr));	
				return addr[i];
			}
		}
// be careful
	if ( count <= 0 )
		return naddr;

	int index = rand() % count;
	recent_aid = aid[index];
	
	//recent_nid = nid[index];
	
	//return recent_nid;
//slee_printf("getrand %d node addr %s\n", aid[index], inet_ntoa(addr[index].sin_addr));	
	return addr[index];
}
/*
void IdCache::getNodeId(int ai, struct sockaddr_in * nid) {

  //	if ( ai >= 0 && recent_aid == ai )
  //		return recent_nid;
 //PRM cache
	unsigned long agent_addr;
  
  nid = NULL;

	if ( ai >= 0 )
		for ( int i = 0; i < count; i++ ) {
			if ( aid[i] == ai ) {
				recent_aid = aid[i];
				//recent_nid = nid[i];				
				//return nid[i];

				nid = &(addr[i]);
				agent_addr=strtoul(inet_ntoa(addr[i].sin_addr), NULL, 10);
				nid->sin_addr.s_addr=addr[i].sin_addr.s_addr;
				return;
			}
		}

	if ( count <= 0 )
		return;

	int index = rand() % count;
	recent_aid = aid[index];
	
	//recent_nid = nid[index];
	
	//return recent_nid;
	nid = &(addr[index]);
	agent_addr=strtoul(inet_ntoa(addr[index].sin_addr), NULL, 10);
	nid->sin_addr.s_addr=addr[index].sin_addr.s_addr;
	return;
}
*/

int IdCache::addId( AppPacket *ap, int in_rscore, int in_mscore ) {
	query_random *ptr = &ap->u.random_p;
//slee_printf("rand %d node addr %s\n", ap->u.random_p.rand_id.agent_id, inet_ntoa(ap->u.random_p.rand_id.agent_addr.sin_addr));	
	//PRM
	//return addId( ptr->rand_id.agent_id, ptr->rand_id.node_id, ptr->level, in_rscore, in_mscore );
	return addId( ptr->rand_id.agent_id, &(ptr->rand_id.agent_addr), ptr->level, in_rscore, in_mscore );
}

//int IdCache::addId(int ai, int ni, int lvl, int in_rscore, int in_mscore ) {
int IdCache::addId(int ai, struct sockaddr_in * ni, int lvl, int in_rscore, int in_mscore){
	int ret = 1;
		   int index = found(ai/*, ni*/);
	if ( index >= 0 ) {
		sum_recv -= recv_score[index];
		recv_score[index] = in_rscore;
		sum_recv += recv_score[index];

		sum_miss -= miss_score[index];
		miss_score[index] = in_mscore;
		sum_miss += miss_score[index];
		ret = 0;
	}
	//not found
	else if ( count < MAX_ID_CACHE ) {
		aid[count] = ai;
		
		//nid[count] = ni;
		//PRM
		memcpy(&(addr[count]), ni, sizeof(struct sockaddr_in));

		recv_score[count] = in_rscore;
		miss_score[count] = in_mscore;
		level[count] = 1 << (2 *lvl);

		sum_level += level[count];
		sum_recv += recv_score[count];
		sum_miss += miss_score[count];
		count++;
	}
	else {
		int index = getId();
		aid[index] = ai;
		//nid[index] = ni;
		//PRM
		memcpy(&(addr[index]), ni, sizeof(struct sockaddr_in));

		sum_level -= level[index];
		sum_recv -= recv_score[index];
		sum_miss -= miss_score[index];

		recv_score[count] = in_rscore;
		miss_score[count] = in_mscore;
		level[index] = 1 << (2 * lvl);

		sum_level += level[index];
		sum_recv += recv_score[index];
		sum_miss += miss_score[index];
	}
	return ret;
}

int IdCache::getId() {

	int i;
	for ( i = 0; i < count; i++ ) {
		if ( recv_score[i] == 0 && miss_score[i] == 0 )
			return i;
	}

	//fprintf(stderr, "non-zero id\n");
	return rand() % count;
}

int IdCache::found( int ai/*, int ni*/) {
	int i;
	for ( i = 0 ; i < count; i++ ) 
	  if ( aid[i] ==  ai /*&& nid[i] == ni */)
			return i;

	return -1;
}

void IdCache::dump() {
/*
	printf("----------------------\n");
	printf("sum_recv: %d, sum_miss: %d\n", sum_recv, sum_miss);
	for ( int i = 0 ; i < count; i++ ) {
	  printf("\t%5d %5d %5d\n", aid[i], recv_score[i], miss_score[i]);
	}
*/
	printf("idCache DUMP: ");
	for ( int i = 0 ; i < count; i ++ ) {
		printf("%d ", aid[i] );
	}
	printf("\n");
}

////////////////////////////////////

void RxmitCache::clear() {
	memset( valid, 0, RXMIT_CACHE_SIZE ); 
}

int RxmitCache::lookup( int seq , double now) {
	int index = seq % RXMIT_CACHE_SIZE;
	if ( valid[index] && seq_no[index] == seq && timeout[index] > now )
		return 1;
	return 0;
}

int RxmitCache::insert( int seq, double rtt ) {

  //slee_printf(" The RTT time here in Retransmit_cache is %d %f\n", seq, rtt);

	assert ( rtt >= 0.0 );
	//	double now = Scheduler::Clock();
	double now = my_clock();

	if ( lookup( seq, now ) )
		return 1;

	int index = seq % RXMIT_CACHE_SIZE;
	valid[index] = 1;
	seq_no[index] = seq;
	timeout[index] = now + rtt * RXMIT_MULT;

	//fprintf(stderr, "%f %f %f\n", now, rtt, timeout[index] );

	return 0;
}
		
	

