/*
 * File: coop.cc
 * Author: Suman Banerjee <suman@cs.umd.edu>
 * Date: 15th February, 2002
 * Terms: GPL
 *
 * NICE Application Layer Multicast
 */

#include <sys/types.h>
#include "o_timeout.h"
#include "coop-agent.h"
#include "talker_utils.h"
#include "timer.h"
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>

int NUM_PACKETS = 3000;
float PACKET_GAP = 1.0;
int START_DELAY = 20;

void SIGINTHandler(int SignalType) {
  printf("Disconnecting from talker...\n");
  exit(disconnect_talker());
}

int main(int argc, char ** argv) {

  struct sigaction handler;

#ifdef LOG_CLUSTER_INFO_DETAILED
  if (argc != 6 && argc != 8) {
    printf("usage: %s <agent id> <port num> <bse hostname> <talker hostname> <talker port> [<drawer hostname> <drawer port>]\n", argv[0]);
    exit(-1);
  }
#else
  if(argc != 6) {
    printf("usage: %s <agent id> <port num> <bse hostname> <talker hostname> <talker port>\n", argv[0]);
    exit(-1);
  }
#endif
  handler.sa_handler = SIGINTHandler;
  if(sigfillset(&handler.sa_mask) < 0) {
    printf("Couldn't fill signal set\n");
    return -1;
  }
  handler.sa_flags = 0;
  if(sigaction(SIGINT, &handler, 0) < 0) {
    printf("Couldn't set signal action\n");
    return -1;
  }
  coopAgent mycoop;
	int port_num = atoi(argv[2]);
#ifdef LOG_CLUSTER_INFO_DETAILED
  if(argc == 8)
    mycoop.init(atoi(argv[1]),0,atoi(argv[2]), argv[3], argv[6], atoi(argv[7]));
  else
    mycoop.init(atoi(argv[1]),0,atoi(argv[2]), argv[3]);
#else
  mycoop.init(atoi(argv[1]),0,atoi(argv[2]),argv[3]);
#endif
//slee - change
/*
	int any_sock = socket( PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if ( any_sock < 0 )
		printf("any socket socket() error\n");

	struct sockaddr_in any_addr;
	memset( &any_addr, 0, sizeof(struct sockaddr_in));
	any_addr.sin_family = AF_INET;
	any_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	any_addr.sin_port = htons(port_num);
	int bindret;
	if ( (bindret = bind( any_sock, (struct sockaddr*) &any_addr, sizeof(struct sockaddr))) < 0 )
		printf("bind to any_sock failed(), %d port: %d\n\n", bindret, port_num);

	mycoop.add_listening_fd(any_sock, NULL);

*/
  mycoop.add_listening_fd(mycoop.udp_sock, NULL);

//end of slee change

  setup_talker(argv[4], atoi(argv[5]));
  initTimeout();
  mycoop.start();
  return 0; 
}
