/*
 * File: coop.cc
 * Author: Suman Banerjee <suman@cs.umd.edu>
 * Date: 15th February, 2002
 * Terms: GPL
 *
 * NICE Application Layer Multicast
 */

#include "o_timeout.h"
#include "coop-agent.h"
#include "talker_utils.h"
#include "timer.h"
#include <stdlib.h>
#include <unistd.h>

int NUM_PACKETS = 3000;
float PACKET_GAP = 1.0;
int START_DELAY = 20;

void SIGINTHandler(int SignalType) {
  printf("Disconnecting from talker...\n");
  exit(disconnect_talker());
}

int main(int argc, char ** argv) {

  struct sigaction handler;

#ifdef LOG_CLUSTER_INFO_DETAILED
  if (argc != 6 && argc != 8) {
    printf("usage: %s <agent id> <port num> <bse hostname> <talker hostname> <talker port> [<drawer hostname> <drawer port>]\n", argv[0]);
    exit(-1);
  }
#else
  if(argc != 6) {
    printf("usage: %s <agent id> <port num> <bse hostname> <talker hostname> <talker port>\n", argv[0]);
    exit(-1);
  }
#endif
  handler.sa_handler = SIGINTHandler;
  if(sigfillset(&handler.sa_mask) < 0) {
    printf("Couldn't fill signal set\n");
    return -1;
  }
  handler.sa_flags = 0;
  if(sigaction(SIGINT, &handler, 0) < 0) {
    printf("Couldn't set signal action\n");
    return -1;
  }
  coopAgent mycoop;
#ifdef LOG_CLUSTER_INFO_DETAILED
  if(argc == 8)
    mycoop.init(atoi(argv[1]),0,atoi(argv[2]), argv[3], argv[6], atoi(argv[7]));
  else
    mycoop.init(atoi(argv[1]),0,atoi(argv[2]), argv[3]);
#else
  mycoop.init(atoi(argv[1]),0,atoi(argv[2]),argv[3]);
#endif
  mycoop.add_listening_fd(mycoop.udp_sock, NULL);
  setup_talker(argv[4], atoi(argv[5]));
  initTimeout();
  mycoop.start();
  return 0; 
}
