#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include "talker_utils.h"

static int sock = -1;
static struct sockaddr_in talker_addr;

//extern struct timeval start_time;
//extern float start_time_usec_float;

extern double my_clock();

#define LOG_TALKER_RESPONSE

double estimate_rtt_wrapper(struct sockaddr_in their_addr) {
  char ping_info[6];
  double ping_time = 40000;
/*
  struct timeval time1;
  gettimeofday(&time1, NULL);
  double d_time = (time1.tv_sec + (((float)time1.tv_usec) / 1000000.0));
*/

  if(sock <= 0) {
    printf("Talker-utils: WARNING: estimate_rtt_wrapper() called but talker location not defined\n");
    // return 40000;
  }
  else {
    memcpy(ping_info, &their_addr.sin_addr.s_addr, sizeof(their_addr.sin_addr.s_addr));
    memcpy(&ping_info[4], &their_addr.sin_port, sizeof(their_addr.sin_port));
    if(send(sock, ping_info, 6, 0) < 0) {
      printf("Talker-utils: Send failed\n");
      // return 40000;
    }
    else if(recv(sock, ping_info, 6, 0) < 0) {
      printf("Talker-utils: Error receiving ping response\n");
       // return 40000;
    }
    else
      ping_time = (float)ntohl(*((long*)ping_info));
  }
#ifdef LOG_TALKER_RESPONSE
    printf("[Talker-resp] from [ %s %d ] %f at %f\n", inet_ntoa(their_addr.sin_addr), their_addr.sin_port, ping_time, my_clock() );
#endif
  return ping_time;
}

int setup_talker(char* hostname, short port) {
  struct hostent* h;
  if((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
    printf("Talker-utils: Could not create socket for talker\n");
    return -1;
  }
  h = gethostbyname(hostname);
  if(h == 0x0) {
    printf("Talker-utils: Could not resolve talker hostname\n");
    return -1;
  }
  memset(&talker_addr, 0, sizeof(talker_addr));
  talker_addr.sin_family = AF_INET;
  talker_addr.sin_addr = *((in_addr*)h->h_addr_list[0]);
  talker_addr.sin_port = htons(port);
  if(connect(sock, (struct sockaddr*)&talker_addr, sizeof(talker_addr)) < 0) {
    printf("Talker-utils: Could not connect to talker\n");
    return -1;
  }
  return 0;
}

int disconnect_talker() {
  char bye[] = "bye";
  int ret;

  if(sock < 0) {
    printf("Talker-utils: WARNING: disconnect_talker() called on an unconnected socket\n");
    return 0;
  }
  if(send(sock, bye, strlen(bye), 0) < 0) {
    printf("Talker-utils: Error disconnecting from talker\n");
  }
  ret = close(sock);
  sock = -1;
  return ret;
}

int talker_connected() {
  return sock >= 0;
}
