/*
 * File: bse.cc
 * Author: Suman Banerjee <suman@cs.umd.edu>
 * Date: 15th February, 2002
 * Terms: GPL
 *
 * NICE Application Layer Multicast
 */

#include "bse-agent.h"
#include "talker_utils.h"
#include "o_timeout.h"

void SIGINTHandler(int signalType) {
  printf("Disconnecting from talker...\n");
  exit(disconnect_talker());
}

int main(int argc, char** argv) {
  
  struct sigaction handler;
#ifdef LOG_CLUSTER_INFO_DETAILED
  if(argc != 3 && argc != 5) {
    printf("Usage: %s <talker hostname> <talker port> [<drawer hostname> <drawer port>]\n", argv[0]);
    exit(-1);
  }
#else
  if(argc != 3) {
    printf("Usage: %s <talker hostname> <talker port>\n", argv[0]);
    exit(-1);
  }
#endif
  handler.sa_handler = SIGINTHandler;
  if(sigfillset(&handler.sa_mask) < 0) {
    printf("Couldn't fill signal set\n");
    return -1;
  }
  handler.sa_flags = 0;
  if(sigaction(SIGINT, &handler, 0) < 0) {
    printf("Couldn't set signal action\n");
    return -1;
  }
  bseAgent mybse;
  if(setup_talker(argv[1], atoi(argv[2])) < 0)
    exit(-1);
#ifdef LOG_CLUSTER_INFO_DETAILED
  if(argc == 5) {
    mybse.init(0,0, argv[3], atoi(argv[4]));
  }
  else
    mybse.init(0,0);
#else
  mybse.init(0,0);
#endif
  initTimeout();
  mybse.start();
  return 0;
}
